/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2015 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2016-2018 by TOPPERS PROJECT Educational Working Group.
 * 
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  $Id: i2cshield.c 2416 2016-02-12 12:05:32Z roi $
 */

/* 
 *  I2Cシールド制御プログラムの本体
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <stdio.h>
#include <string.h>
#include <target_syssvc.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include "device.h"
#include "pinmode.h"
#include "i2cshield.h"

/*
 *  サービスコールのエラーのログ出力
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))


#define RXBUFFERSIZE       8

/*
 *  スレーブアドレス定義
 */
#define LCD_ADDR           (0x3E<<1)
//#define EEPROM_ADDR        (0x50<<1)
#define ADT7410_ADDR       (0x48<<1)

#define ADT7410_TEMP_H	0x00
#define ADT7410_TEMP_L	0x01
#define ADT7410_STATUS	0x02

#define LCD_locate(h, x, y) aqm0802_set_command((h), (0x80 + ((y) << 6) + (x)))
#define LCD_puts(h, b)      aqm0802_set_data((h), (b), strlen((const char *)(b)))


/* Buffer used for reception */
uint8_t aRxBuffer[RXBUFFERSIZE];

uint8_t lcd_buf[2][12];

/*
 *  SW1割込み
 */
void sw_int(void)
{
	syslog_0(LOG_NOTICE, "## sw_int() ##");
}

static void
set_value(uint8_t *buf, int value)
{
    buf[1] = (value % 10) + '0';
    buf[0] = (value / 10) + '0';
}

static void
set_time(uint8_t *buf, long time)
{
    int wk = time/1000;
    set_value(&buf[6], wk % 60);
    wk = wk / 60;
    buf[5] = ':';
    set_value(&buf[3], wk % 60);
    wk = wk / 60;
    buf[2] = ':';
    set_value(&buf[0], wk);
}


/*
 *  メインタスク
 */
void
main_task(intptr_t exinf)
{
	FPGAI2C_Handle_t *hi2c;
	CLCD_Handler_t  CLcdHandle;
	ER_UINT	ercd;
	SYSTIM  tim;
	uint32_t temp;
	uint32_t count = 0;

	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG)));
	syslog(LOG_NOTICE, "Sample program starts (exinf = %d).", (int_t) exinf);

	/*
	 *  シリアルポートの初期化
	 *
	 *  システムログタスクと同じシリアルポートを使う場合など，シリアル
	 *  ポートがオープン済みの場合にはここでE_OBJエラーになるが，支障は
	 *  ない．
	 */
	ercd = serial_opn_por(TASK_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(TASK_PORTID,
							(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));

	syslog_0(LOG_NOTICE, "I2C SEND/RECV START");

	if((hi2c = fpgai2c_init(FPGAI2C1_PORTID, 100000)) == NULL){
		/* Initialization Error */
		syslog_0(LOG_ERROR, "## I2C ERROR(1) ##");
	}
	hi2c->semlock = I2CLOC_SEM;

	syslog_0(LOG_NOTICE, "LCD INITAILIZE !");
	memset(&lcd_buf[0][0], ' ', 8);
	memset(&lcd_buf[1][0], ' ', 8);

	CLcdHandle.hi2c  = hi2c;
	CLcdHandle.saddr = LCD_ADDR;
	if((ercd = aqm0802_init(&CLcdHandle)) != E_OK){
		syslog_1(LOG_ERROR, "## LCD INIT ERROR(%d)[%08x] ##", ercd);
		goto stop_task;
	}

	get_tim(&tim);
	set_time(&lcd_buf[0][0], tim);
	LCD_locate(&CLcdHandle, 0, 0);
	LCD_puts(&CLcdHandle, &lcd_buf[0][0]);

#ifdef EEPROM_ADDR
	syslog_1(LOG_NOTICE, "READ EEPROM [%08x] !", aRxBuffer);
	if((ercd = fpgai2c_memread(hi2c, (uint16_t)EEPROM_ADDR, 0, 2, (uint8_t*)aRxBuffer, RXBUFFERSIZE)) != E_OK){
	    /* I2C RECEIVE ERROR */
		syslog_1(LOG_ERROR, "## I2C ERROR(12)(%d) ##", ercd);
		goto stop_task;
	}
#endif

	syslog_0(LOG_NOTICE, "TEMP SENSER !");
	do{
		/* STATUS READ */
		if((ercd = fpgai2c_memread(hi2c, (uint16_t)ADT7410_ADDR, ADT7410_STATUS, 1, (uint8_t*)aRxBuffer, 1)) != E_OK){
			/* I2C RECEIVE ERROR */
			syslog_1(LOG_ERROR, "## I2C ERROR(13)(%d) ##", ercd);
			goto stop_task;
		}
		temp = aRxBuffer[0];
#if 1	/* ROI DEBUG */
		syslog_1(LOG_NOTICE, "## temp[%02x] ##", temp);
		dly_tsk(500);
		break;
#endif	/* ROI DEBUG */
	}while((temp & 0x80) != 0);

	while(count < 1000){
		if((ercd = fpgai2c_memread(hi2c, (uint16_t)ADT7410_ADDR, ADT7410_TEMP_H, 1, (uint8_t*)aRxBuffer, 1)) != E_OK){
			/* I2C RECEIVE ERROR */
			syslog_1(LOG_ERROR, "## I2C ERROR(14)(%d) ##", ercd);
			goto stop_task;
		}
		temp = aRxBuffer[0] << 8;
		if((ercd = fpgai2c_memread(hi2c, (uint16_t)ADT7410_ADDR, ADT7410_TEMP_L, 1, (uint8_t*)aRxBuffer, 1)) != E_OK){
			/* I2C RECEIVE ERROR */
			syslog_1(LOG_ERROR, "## I2C ERROR(15)(%d) ##", ercd);
			goto stop_task;
		}
		temp += aRxBuffer[0];
		temp /= 13;

		get_tim(&tim);
		set_time(&lcd_buf[0][0], tim);
		LCD_locate(&CLcdHandle, 0, 0);
		LCD_puts(&CLcdHandle, &lcd_buf[0][0]);
		set_value(&lcd_buf[1][0], temp/10);
		lcd_buf[1][2] = '.';
		lcd_buf[1][3] = (temp % 10) + '0';
		LCD_locate(&CLcdHandle, 0, 1);
		LCD_puts(&CLcdHandle, &lcd_buf[1][0]);
		if((count % 10) == 0)
			syslog_2(LOG_NOTICE, "count(%d) temp(%d)", count, temp);
		dly_tsk(1000);
		count++;
	}

stop_task:
	syslog_0(LOG_NOTICE, "## STOP ##");
	slp_tsk();
	syslog(LOG_NOTICE, "Sample program ends.");
	SVC_PERROR(ext_ker());
}

