/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2015 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2016-2019 by TOPPERS PROJECT Educational Working Group.
 * 
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  $Id: lcdshield.c 2416 2019-04-09 18:05:10Z roi $
 */

/* 
 *  LCDシールド制御プログラムの本体
 *  漢字フォントを選択したい場合、make FONT_MODE=1でビルドする
 *  漢字フォントデータは、以下のフォントファイルを使用する
 *  使用時はフォントファイルをSDカードのrootディレクトリに置いて
 *  ください
 *  shinonome_font12.fnt(12x12)
 *  shinonome_font16.fnt(16x16)
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <stdio.h>
#include <string.h>
#include <target_syssvc.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include "device.h"
#include "pinmode.h"
#include "storagedevice.h"
#include "fcntl.h"
#include "lcdshield.h"
#include "topamelogo.h"

/*
 *  サービスコールのエラーのログ出力
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))


LCD_Handler_t   LcdHandle;
LCD_DrawProp_t DrawProp;

static Point testPolygon[7] = {
	{0, 0},
	{239, 319},
	{120, 0},
	{120, 319},
	{239, 0},
	{0, 319},
	{0, 0}
};

static const uint8_t string1[] = {
#if FONT_MODE==1
	0xE3,0x83,0x8F,0xE3,0x83,0xAD,0xE3,0x83,0xBC,0xE4,0xB8,0x96, 0xE7,0x95,0x8C,' ', '!', 0
#else
	'H', 'e', 'l', 'l', 'o', ' ', 'W', 'o', 'r', 'l', 'd', ' ', '!', 0
#endif
};

/*
 *  ユーザー領域設定
 */
static volatile int  vol_status = 0;

/*
 *  SW1割込み
 */
void sw_int(void)
{
	syslog_0(LOG_NOTICE, "## sw_int() ##");
}

/*
 *  ストレージマネージャからのコールバック
 */
static void
storagedev_notice(void *psdev, bool_t sw)
{
	if(sw)
		vol_status |= 1<<((StorageDevice_t*)psdev)->_sdev_devno;
	else
		vol_status &= ~(1<<((StorageDevice_t*)psdev)->_sdev_devno);
	syslog_3(LOG_NOTICE, "MEDIA[%08x][%04x] on/off(%d) !", psdev, vol_status, sw);
}

/*
 *  FAT時間の取得
 */
static uint32_t
get_fat_time(void)
{
	int    reqtime = 0;
#ifndef SAMPLE_PROGRAM
	struct tm2 timedate;

	rtc_get_time(&timedate);
	reqtime  = ((timedate.tm_year-10)<<25)+(timedate.tm_mon<<21)+(timedate.tm_mday<<16);
	reqtime += (timedate.tm_hour<<11)+(timedate.tm_min<<5)+(timedate.tm_sec/2);
#else
	reqtime = 0x00210000;
#endif
	return reqtime;
}

/*
 *  グラフィック表示テスト
 */
void
grapics_test(LCD_Handler_t *hlcd)
{
	uint16_t x, y;
	int      i;

	DrawProp.BackColor = ILI9341_BLACK;
	lcd_fillScreen(&DrawProp);
	dly_tsk(1000);
	lcd_drawPixel(hlcd, hlcd->_width/2, hlcd->_height/2, ILI9341_GREEN);
	dly_tsk(1000);

	lcd_fillScreen(&DrawProp);
	for(y=0; y < hlcd->_height; y+=5) {
		lcd_drawFastHLine(hlcd, 0, y, hlcd->_width, ILI9341_RED);
	}
	for (x=0; x < hlcd->_width; x+=5) {
    	lcd_drawFastVLine(hlcd, x, 0, hlcd->_height, ILI9341_BLUE);
	}
	dly_tsk(1000);

	lcd_fillScreen(&DrawProp);
	DrawProp.TextColor = ILI9341_GREEN;
	lcd_drawPolygon(&DrawProp, testPolygon, sizeof(testPolygon)/sizeof(Point));
	dly_tsk(1000);

	lcd_fillScreen(&DrawProp);
	DrawProp.TextColor = ILI9341_GREEN;
	for (x=0; x < hlcd->_width; x+=6) {
		lcd_drawRect(&DrawProp, hlcd->_width/2 -x/2, hlcd->_height/2 -x/2 , x, x);
	}
	for(i = 0 ; i < 3 ; i++){
		lcd_invertDisplay(hlcd, true);
		dly_tsk(500);
		lcd_invertDisplay(hlcd, false);
		dly_tsk(500);
	}
	lcd_fillScreen(&DrawProp);
	lcd_drawBitmap(hlcd, 33, 48, (uint8_t *)topamelogo);
	dly_tsk(1000);
}

/*
 *  メインタスク
 */
void
main_task(intptr_t exinf)
{
	StorageDevice_t *psdev;
	LCD_Handler_t   *hlcd;
	FPGASPI_Handle_t  *hspi;
	ER_UINT	ercd;
	int  i;

	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG)));
	syslog(LOG_NOTICE, "Sample program starts (exinf = %d).", (int_t) exinf);

	/*
	 *  シリアルポートの初期化
	 *
	 *  システムログタスクと同じシリアルポートを使う場合など，シリアル
	 *  ポートがオープン済みの場合にはここでE_OBJエラーになるが，支障は
	 *  ない．
	 */
	ercd = serial_opn_por(TASK_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(TASK_PORTID,
							(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));

	/*
	 *  SDMMCデバイスの設定
	 */
	psdev = SDMGetStorageDevice(0);
	psdev->_sdev_notice = storagedev_notice;
	SDeviceHead._get_datetime = get_fat_time;

	syslog_0(LOG_NOTICE, "## START ##");
#ifdef RST_PORT
	pinMode(RST_PORT, OUTPUT);
#endif
	pinMode(DC_PORT, OUTPUT);
	hspi = fpgaspi_init(FPGASPI1_PORTID);
	if(hspi == NULL){
		syslog_0(LOG_ERROR, "hspi initial error !");
		slp_tsk();
	}
	hspi->semid   = SPITRS_SEM1;
	hspi->semlock = 0;

	hlcd = &LcdHandle;
	hlcd->hspi     = hspi;
	hlcd->spi_lock = 0;
	hlcd->cs_pin   = CS_PORT;
	hlcd->dc_pin   = DC_PORT;
#ifdef RST_PORT
	hlcd->rst_pin  = RST_PORT;
#else
	hlcd->rst_pin  = 32;
#endif
	DrawProp.hlcd  = hlcd;
	lcd_init(hlcd, 0);

	DrawProp.BackColor = ILI9341_BLACK;
	DrawProp.TextColor = ILI9341_WHITE;
	grapics_test(hlcd);

	DrawProp.TextColor = ILI9341_BLACK;
	DrawProp.BackColor = ILI9341_WHITE;
	lcd_fillScreen(&DrawProp);
	lcd_fillRect(hlcd, 0, 0, hlcd->_width, hlcd->_height/3, ILI9341_BLUE);

	DrawProp.TextColor = ILI9341_BLACK;
	DrawProp.BackColor = ILI9341_WHITE;

	DrawProp.TextColor = ILI9341_WHITE;
	DrawProp.BackColor = ILI9341_BLUE;
	DrawProp.pFont = &Font16;
	lcd_DisplayStringAt(&DrawProp, 0, 12, (uint8_t *)"TOPPERS", CENTER_MODE);
	lcd_DisplayStringAt(&DrawProp, 0, 32, (uint8_t *)"BASE PROTFORM(CV)", CENTER_MODE);

	DrawProp.TextColor = ILI9341_BLACK;
	DrawProp.BackColor = ILI9341_WHITE;
	DrawProp.pFont = &Font16;
	lcd_DisplayStringAt(&DrawProp, 0, 120, (uint8_t *)string1, CENTER_MODE);
	DrawProp.pFont = &Font12;
	lcd_DisplayStringAt(&DrawProp, 0, 140, (uint8_t *)string1, CENTER_MODE);
	DrawProp.pFont = &Embedded_Font9;
	lcd_DisplayStringAt(&DrawProp, 0, 160, (uint8_t *)"Helle World !", CENTER_MODE);

	DrawProp.TextColor = ILI9341_RED;
	for(i = 20 ; i < 40 ; i++){
		lcd_DrawCircle(&DrawProp, hlcd->_width/2, hlcd->_height/2+30, i);
		if(i == 25)
			i += 10;
	}

	syslog_0(LOG_NOTICE, "LCD TEST END");
	syslog_0(LOG_NOTICE, "## STOP ##");
	slp_tsk();
	fpgaspi_deinit(hspi);

	syslog(LOG_NOTICE, "Sample program ends.");
//	SVC_PERROR(ext_ker());
}

