/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2015 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2016-2017 by TOPPERS PROJECT Educational Working Group.
 * 
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  @(#) $Id: i2c_demo.c 64 2017-05-30 18:19:15Z roi $
 */

/* 
 *  I2C_DEMOプログラムの本体
 *
 *  DK-DEV-5CGTD9Nボード用のI2Cテストプログラム．
 *
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <sil.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include "device.h"
#include "socfpga_cv.h"
#include "i2c.h"
#include "i2c_demo.h"

/*
 *  サービスコールのエラーのログ出力
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))


#define ADXL345_ADDR        (0x1D<<1)

/*
 *  シリアルポート番号のテーブル
 */
uint_t const serial_port[TNUM_PRCID] = {
#ifndef G_SYSLOG
	TASK_PORTID_PRC1,
#else /* G_SYSLOG */
	TASK_PORTID_G_SYSLOG,
#endif /* G_SYSLOG */
    
#if TNUM_PRCID >= 2
    TASK_PORTID_PRC2,
#endif /* TNUM_PRCID >= 2 */
};

uint8_t rxbuffer[16];
uint8_t txbuffer[16];
uint8_t dvaddr[8];

/*
 *  タスクのマイグレーション先のプロセッサの指定
 */
ID  tsk_mig_prc;


/*
 *  CPU例外ハンドラ
 */
#ifdef CPUEXC1

void
cpuexc_handler(void *p_excinf)
{
	ID		tskid;

	syslog(LOG_NOTICE, "CPU exception handler1 (p_excinf = %08p).", p_excinf);
	if (sns_ctx() != true) {
		syslog(LOG_WARNING,
					"sns_ctx() is not true in CPU exception handler.");
	}
	if (sns_dpn() != true) {
		syslog(LOG_WARNING,
					"sns_dpn() is not true in CPU exception handler.");
	}
	syslog(LOG_INFO, "sns_loc = %d sns_dsp = %d sns_tex = %d",
									sns_loc(), sns_dsp(), sns_tex());

	syslog(LOG_INFO, "xsns_dpn = %d xsns_xpn = %d",
									xsns_dpn(p_excinf), xsns_xpn(p_excinf));

    
	if (xsns_xpn(p_excinf)) {
		syslog(LOG_NOTICE, "Sample program ends with exception.");
		SVC_PERROR(ext_ker());
		assert(0);
	}

	SVC_PERROR(iget_tid(&tskid));
	SVC_PERROR(iras_tex(tskid, 0x8000U));
}

#endif /* CPUEXC1 */

static uint_t counter;


void
cyclic_handler(intptr_t exinf)
{
	wdog_reset(WDOG1_PORTID);
}

/*
 *  グローバル初期化ルーチン
 */
void
global_inirtn(intptr_t exinf)
{
	counter = 1;
}

/*
 *  グローバル終了ルーチン
 */
void
global_terrtn(intptr_t exinf)
{
	syslog(LOG_EMERG, "global_terrtn exinf = %d, counter = %d", exinf, counter++);
}

/*
 *  ローカル初期化ルーチン
 */
void
local_inirtn(intptr_t exinf)
{
	syslog(LOG_NOTICE, "local_inirtn exinf = %d, counter = %d", exinf, counter++);
}

/*
 *  ローカル終了ルーチン
 */
void
local_terrtn(intptr_t exinf)
{
	syslog(LOG_NOTICE, "local_terrtn exinf = %d, counter = %d", exinf, counter++);
}

#if 1	/* ROI DEBUG */
void sw_int(void)
{
	syslog_0(LOG_NOTICE, "## sw_int() ##");
}
#endif	/* ROI DEBUG */

/******************************************************************************/
/*!
 * Delay function
 *
 * \param      us delay interval
 */
void delay_us(uint32_t us)
{
	while(us != 0){
		wdog_reset(WDOG1_PORTID);
		sil_dly_nse(1000);
		us--;
	}
}

/**
  * @brief  Memory Tx Transfer completed callbacks.
  * @param  hi2c: pointer to a I2C_Handle_t structure that contains
  *         the configuration information for I2C module
  * @retval None
  */
void HAL_I2C_MemTxCpltCallback(I2C_Handle_t *hi2c)
{
  syslog_3(LOG_NOTICE, "## HAL_I2C_MemTxCpltCallback[%08x] Size(%d) Count(%d) ##", hi2c, hi2c->XferSize, hi2c->XferCount);
  /* NOTE : This function Should not be modified, when the callback is needed,
            the HAL_I2C_TxCpltCallback could be implemented in the user file
   */
}

/**
  * @brief  Memory Rx Transfer completed callbacks.
  * @param  hi2c: pointer to a I2C_Handle_t structure that contains
  *         the configuration information for I2C module
  * @retval None
  */
void HAL_I2C_MemRxCpltCallback(I2C_Handle_t *hi2c)
{
  syslog_3(LOG_NOTICE, "## HAL_I2C_MemRxCpltCallback[%08x] Size(%d) Count(%d) ##", hi2c, hi2c->XferSize, hi2c->XferCount);
  /* NOTE : This function Should not be modified, when the callback is needed,
            the HAL_I2C_TxCpltCallback could be implemented in the user file
   */
}

/**
  * @brief  I2C error callbacks.
  * @param  hi2c: pointer to a I2C_Handle_t structure that contains
  *         the configuration information for I2C module
  * @retval None
  */
void HAL_I2C_ErrorCallback(I2C_Handle_t *hi2c)
{
  syslog_2(LOG_NOTICE, "## HAL_I2C_ErrorCallback(%08x) Error[%08x] ##", hi2c, hi2c->ErrorCode);
  /* NOTE : This function Should not be modified, when the callback is needed,
            the HAL_I2C_ErrorCallback could be implemented in the user file
   */
}


/*
 *  メインタスク
 */
void main_task(intptr_t exinf)
{
	ER status = E_OK;
	ER_UINT	ercd;
	SYSTIM tim;
	int random_seed;

	tsk_mig_prc = 1;

	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_NOTICE), LOG_UPTO(LOG_EMERG)));
	syslog(LOG_NOTICE, "Sample program starts (exinf = %d).", (int_t) exinf);

	/*
	 *  シリアルポートの初期化
	 *
	 *  システムログタスクと同じシリアルポートを使う場合など，シリアル
	 *  ポートがオープン済みの場合にはここでE_OBJエラーになるが，支障は
	 *  ない．
	 */
	ercd = serial_opn_por(((int_t)serial_port[exinf-1]));
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(serial_port[exinf-1],
							(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));


    // System init
	get_tim(&tim);
	random_seed = tim;
	syslog_1(LOG_NOTICE, "INFO: Using random seed = 0x%04x.", random_seed);
	srand(random_seed);

    // Demo LCD
    if(status == E_OK)
    {
        ercd = i2c_demo_master_lcd();
		status = ercd;
    }
	dly_tsk(1000);

    // Demo EEPROM
    if(status == E_OK)
    {
        ercd = i2c_demo_master_eeprom();
		status = ercd;
    }
	dly_tsk(1000);

    // Demo master and slave
    if(status == E_OK)
    {
        ercd = i2c_demo_master_slave();
		status = ercd;
    }

    // Report status
    if (status == E_OK)
    {
        syslog_0(LOG_NOTICE, "RESULT: All tests successful.\n");
    }
    else
    {
        syslog_0(LOG_NOTICE, "RESULT: Some failures detected.\n");
    }

	syslog_0(LOG_NOTICE, "## STOP ##");
	slp_tsk();

	syslog(LOG_NOTICE, "Sample program ends.");
	SVC_PERROR(ext_ker());
	assert(0);
}
