/******************************************************************************
*
* Copyright 2014 Altera Corporation. All Rights Reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice,
* this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above copyright notice,
* this list of conditions and the following disclaimer in the documentation
* and/or other materials provided with the distribution.
*
* 3. The name of the author may not be used to endorse or promote products
* derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER "AS IS" AND ANY EXPRESS OR
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, ARE DISCLAIMED. IN NO
* EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
* OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
* OF SUCH DAMAGE.
*
******************************************************************************/

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <sil.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include <stdio.h>
#include <string.h>

#include "device.h"
#include "i2c.h"
#include "i2c_demo.h"

#define I2C_EEPROM_ADDRESS                  (0xA2)    // I2C address of EEPROM module
#define I2C_EEPROM_SPEED                    100000         // I2C bus speed for accessing EEPROM module
#define I2C_EEPROM_BLOCK_SIZE               4096           // EEPROM module size in bytes
#define I2C_EEPROM_PAGE_SIZE                32             // EEPROM page size in bytes
#define I2C_EEPROP_WRITE_TIME_US            5000           // Maximum EEPROM write duration

extern void HAL_I2C_MemTxCpltCallback(I2C_Handle_t *hi2c);
extern void HAL_I2C_MemRxCpltCallback(I2C_Handle_t *hi2c);
extern void HAL_I2C_ErrorCallback(I2C_Handle_t *hi2c);

// Write to EEPROM
static ER eeprom_write(I2C_Handle_t * device, uint16_t address, void * buffer, uint16_t size);

// Demonstrate EEPROM
static ER eeprom_demo(I2C_Handle_t * device);


/******************************************************************************/
/*!
 * Generic EEPROM write function.
 *
 * \param       device I2C device.
 * \param         address where to write the data
 * \param         data bytes to be written
 * \param         size how many bytes to be written
 * \return      result of the function
 */
static ER eeprom_write(I2C_Handle_t * device, uint16_t address, void * data, uint16_t size)
{
    ER status = E_OK;
    uint16_t crt_address = address;
    uint8_t * crt_data = (uint8_t*)data;
    uint16_t crt_size = size;

    // Write single bytes until we get to page boundary
    while((status == E_OK) &&
          ((crt_address % I2C_EEPROM_PAGE_SIZE) != 0) &&
          (crt_size > 0))
    {
        status = i2c_memwrite(device, I2C_EEPROM_ADDRESS,
										crt_address, 2,
                                         crt_data,
                                         1);
        crt_size--;
        crt_data++;
        crt_address++;
	    // Wait for completion of operation
    	if(status == E_OK)
    	{
        	delay_us(I2C_EEPROP_WRITE_TIME_US);
	    }
    }

    // Write whole pages
    while((status == E_OK) &&
          ((crt_address % I2C_EEPROM_PAGE_SIZE) == 0) &&
          (crt_size >= I2C_EEPROM_PAGE_SIZE))
    {
	    syslog_1(LOG_NOTICE, "INFO: Writing 32 byte page at address = 0x%04x.", (int)crt_address);
        status = i2c_memwrite(device, I2C_EEPROM_ADDRESS,
										crt_address, 2,
                                         crt_data,
                                         I2C_EEPROM_PAGE_SIZE);
    // Wait for completion of operation
    if(status == E_OK)
    {
        delay_us(I2C_EEPROP_WRITE_TIME_US);
    }
        crt_size -= I2C_EEPROM_PAGE_SIZE;
        crt_data += I2C_EEPROM_PAGE_SIZE;
        crt_address += I2C_EEPROM_PAGE_SIZE;
    }

    // Write leftover single bytes at the end
    while((status == E_OK) && (crt_size > 0))
    {
        status = i2c_memwrite(device, I2C_EEPROM_ADDRESS,
										crt_address, 2,
                                         crt_data,
                                         1);
        crt_size--;
        crt_data++;
        crt_address++;
	    // Wait for completion of operation
    	if(status == E_OK)
    	{
        	delay_us(I2C_EEPROP_WRITE_TIME_US);
	    }
    }

    return status;
}

/******************************************************************************/
/*!
 * EEPROM demo function.
 *
 * \param       device I2C device.
 * \return      result of the function
 */
static ER eeprom_demo(I2C_Handle_t * device)
{
    ER status = E_OK;
    static uint8_t read_buffer[I2C_EEPROM_BLOCK_SIZE];
    static uint8_t write_buffer[I2C_EEPROM_BLOCK_SIZE];
    uint16_t test_size;
    uint16_t test_address;
	int i;

    // Generate random address and size
    test_address = rand() % I2C_EEPROM_BLOCK_SIZE;
    test_size = rand() % (I2C_EEPROM_BLOCK_SIZE - test_address) + 1;

    syslog_2(LOG_NOTICE, "INFO: Using random address = 0x%04x, size = %d bytes.", (int)test_address, (int)test_size);

    // Fill the write buffer with random data
    for(i=0; i<test_size; i++)
    {
        write_buffer[i] = (uint8_t)rand();
    }
	dly_tsk(200);

    // Write to EEPROM
    if(status == E_OK)
    {
        syslog_2(LOG_NOTICE, "INFO: Writing to EEPROM address 0x%04x, size = %d bytes.", (int)test_address, (int)test_size);
        status = eeprom_write(device, test_address, write_buffer, test_size);
    }

	dly_tsk(200);
    // Read from EEPROM
    if(status == E_OK)
    {
        syslog_2(LOG_NOTICE, "INFO: Reading from EEPROM address 0x%04x, size = %d bytes.", (int)test_address, (int)test_size);
        memset(read_buffer, 0, test_size);
        syslog_1(LOG_NOTICE, "INFO: Sequentially reading %d bytes from EEPROM.", (int)test_size);
        status = i2c_memread(device, I2C_EEPROM_ADDRESS,
										test_address, 2,
                                        read_buffer,
                                        test_size);
    }

	dly_tsk(200);
    // Compare the results
    if(status == E_OK)
    {
		int i;
        syslog_0(LOG_NOTICE, "INFO: Comparing written data with read back data.");
		syslog_3(LOG_NOTICE, "## read_buffer[%08x] write_buffer[%08x] test_size[%08x] ##", read_buffer, write_buffer, test_size);
		for(i = 0 ; i < test_size ; i++){
			if(read_buffer[i] != write_buffer[i]){
				syslog_3(LOG_NOTICE, "## Comp ERR (%d) r[%02x] w[%02x] ##", i, read_buffer[i], write_buffer[i]);
				dly_tsk(100);
				status = E_OBJ;
			}
        }
    }

    return status;
}

/******************************************************************************/
/*!
 * EEPROM demo wrapper.
 *
 * \return      result of the function
 */
ER i2c_demo_master_eeprom(void)
{
    ER status = E_OK;
    I2C_Handle_t *device;
    I2C_Init_t cfg;

    syslog_0(LOG_NOTICE, "INFO: I2C Demo Master EEPROM started.");

    cfg.AddressingMode  = I2C_ADDRESSINGMODE_7BIT;
	cfg.ClockSpeed      = I2C_EEPROM_SPEED;
	cfg.OwnAddress1     = 0;
	cfg.GeneralCallMode = 0;
	cfg.SdaHoldClock    = 1;
    cfg.restart_enable = I2C_RESATRT_ENABLE;
    cfg.fs_spklen = 2;
	cfg.semid     = I2CTRS1_SEM;
	cfg.semlock   = I2CLOC1_SEM;

    // Init I2C module
    device = i2c_init(I2C1_PORTID, &cfg);
	if(device == NULL)
		status = E_PAR;
	else{
		device->writecallback = HAL_I2C_MemTxCpltCallback;
		device->readcallback  = HAL_I2C_MemRxCpltCallback;
		device->errorcallback = HAL_I2C_ErrorCallback;
	}

    // Demonstrate LCD operation
    if(status == E_OK)
    {
        status = eeprom_demo(device);
    }

    // I2C Cleanup
    if(status == E_OK)
    {
		i2c_deinit(device);
    }

    // Display status
    if(status == E_OK)
    {
        syslog_0(LOG_NOTICE, "INFO: I2C Demo Master EEPROM succeeded.\n");
    }
    else
    {
        syslog_0(LOG_NOTICE, "INFO: I2C Demo Master EEPROM failed.\n");
    }
	return status;
}
