/******************************************************************************
*
* Copyright 2014 Altera Corporation. All Rights Reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice,
* this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above copyright notice,
* this list of conditions and the following disclaimer in the documentation
* and/or other materials provided with the distribution.
*
* 3. The name of the author may not be used to endorse or promote products
* derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER "AS IS" AND ANY EXPRESS OR
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, ARE DISCLAIMED. IN NO
* EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
* OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
* OF SUCH DAMAGE.
*
******************************************************************************/

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <sil.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include <stdio.h>
#include <string.h>

#include "device.h"
#include "i2c.h"
#include "i2c_demo.h"

#define LCD_I2C_ADDRESS      (0x50 >> 1)    // I2C address of LCD module
#define LCD_I2C_SPEED        40000          // I2C bus speed for accessing LCD module
#define LCD_ESCAPE_CHAR      0xfe           // Escape character used to prefix commands
#define LCD_POS_1ST_LINE     0x00           // LCD cursor position for 1st line
#define LCD_POS_2ND_LINE     0x40           // LCD cursor position for 2nd line
#define LCD_PRINT_DELAY_US   500            // Delay in us after printing text on LCD

#define ALT_I2C_SS_DEFAULT_SPKLEN   11

// Commands supported by the LCD display
typedef enum
{
    LCD_COMMAND_DISPLAY_ON = 0,
    LCD_COMMAND_DISPLAY_OFF,
    LCD_COMMAND_SET_CURSOR,
    LCD_COMMAND_CURSOR_HOME,
    LCD_COMMAND_UNDERLINE_CURSOR_ON,
    LCD_COMMAND_UNDERLINE_CURSOR_OFF,
    LCD_COMMAND_MOVE_CURSOR_LEFT_ONE_PLACE,
    LCD_COMMAND_MOVE_CURSOR_RIGHT_ONE_PLACE,
    LCD_COMMAND_BLINKING_CURSOR_ON,
    LCD_COMMAND_BLINKING_CURSOR_OFF,
    LCD_COMMAND_BACKSPACE,
    LCD_COMMAND_CLEAR_SCREEN,
    LCD_COMMAND_SET_CONTRAST,
    LCD_COMMAND_SET_BACKLIGHT_BRIGHTNESS,
    LCD_COMMAND_LOAD_CUSTOM_CHARACTER,
    LCD_COMMAND_MOVE_DISPLAY_ONE_PLACE_TO_THE_LEFT,
    LCD_COMMAND_MOVE_DISPLAY_ONE_PLACE_TO_THE_RIGHT,
    LCD_COMMAND_CHANGE_RS_232_BAUD_RATE,
    LCD_COMMAND_CHANGE_I2C_ADDRESS,
    LCD_COMMAND_DISPLAY_FIRMWARE_VERSION_NUMBER,
    LCD_COMMAND_DISPLAY_RS_232_BAUD_RATE,
    LCD_COMMAND_DISPLAY_I2C_ADDRESS,
} LCD_COMMAND_T;

// Command description
typedef struct
{
    uint8_t command_code;
    uint8_t no_parameters;
    uint16_t execution_time_us;
} LCD_COMMAND_DESC_T;


// Descriptions for all supported commands
static LCD_COMMAND_DESC_T lcd_command_descriptions[] =
{
    {0x41, 0, 100},  // LCD_COMMAND_DISPLAY_ON,
    {0x42, 0, 100},  // LCD_COMMAND_DISPLAY_OFF,
    {0x45, 1, 100},  // LCD_COMMAND_SET_CURSOR,
    {0x46, 0, 1500}, // LCD_COMMAND_CURSOR_HOME,
    {0x47, 0, 1500}, // LCD_COMMAND_UNDERLINE_CURSOR_ON,
    {0x48, 0, 1500}, // LCD_COMMAND_UNDERLINE_CURSOR_OFF,
    {0x49, 0, 100},  // LCD_COMMAND_MOVE_CURSOR_LEFT_ONE_PLACE,
    {0x4A, 0, 100},  // LCD_COMMAND_MOVE_CURSOR_RIGHT_ONE_PLACE,
    {0x4B, 0, 100},  // LCD_COMMAND_BLINKING_CURSOR_ON,
    {0x4C, 0, 100},  // LCD_COMMAND_BLINKING_CURSOR_OFF,
    {0x4E, 0, 100},  // LCD_COMMAND_BACKSPACE,
    {0x51, 0, 1500}, // LCD_COMMAND_CLEAR_SCREEN,
    {0x52, 1, 500},  // LCD_COMMAND_SET_CONTRAST,
    {0x53, 1, 100},  // LCD_COMMAND_SET_BACKLIGHT_BRIGHTNESS,
    {0x54, 9, 200},  // LCD_COMMAND_LOAD_CUSTOM_CHARACTER,
    {0x55, 0, 100},  // LCD_COMMAND_MOVE_DISPLAY_ONE_PLACE_TO_THE_LEFT,
    {0x56, 0, 100},  // LCD_COMMAND_MOVE_DISPLAY_ONE_PLACE_TO_THE_RIGHT,
    {0x61, 1, 3000}, // LCD_COMMAND_CHANGE_RS_232_BAUD_RATE,
    {0x62, 1, 3000}, // LCD_COMMAND_CHANGE_I2C_ADDRESS,
    {0x70, 0, 4000}, // LCD_COMMAND_DISPLAY_FIRMWARE_VERSION_NUMBER,
    {0x71, 0, 10000},// LCD_COMMAND_DISPLAY_RS_232_BAUD_RATE,
    {0x72, 0, 4000}, // LCD_COMMAND_DISPLAY_I2C_ADDRESS,
};

extern void HAL_I2C_MemTxCpltCallback(I2C_Handle_t *hi2c);
extern void HAL_I2C_MemRxCpltCallback(I2C_Handle_t *hi2c);
extern void HAL_I2C_ErrorCallback(I2C_Handle_t *hi2c);

// Send Command to LCD
static ER lcd_send_command(I2C_Handle_t * device, LCD_COMMAND_T command, uint8_t * params);

// Print text to LCD
static ER lcd_print_text(I2C_Handle_t * device, char * text);

// Demonstrate LCD
static ER lcd_demo(I2C_Handle_t * device);

/******************************************************************************/
/*!
 * Send command to LCD display
 *
 * \param       device  I2C device.
 * \param       command opcode of command to be sent
 * \param       params command parameters or NULL for no parameters
 * \return      result of the function
 */
static ER lcd_send_command(I2C_Handle_t * device, LCD_COMMAND_T command, uint8_t * params)
{
    ER status = E_OK;
    LCD_COMMAND_DESC_T command_description = lcd_command_descriptions[(int)command];
    uint8_t data[10];
    uint8_t data_size = 0;
	int i;

    data[data_size++] = LCD_ESCAPE_CHAR;
    data[data_size++] = command_description.command_code;
    for(i=0; i<command_description.no_parameters; i++)
    {
        data[data_size++] = params[i];
    }

    status = i2c_memwrite(device, LCD_I2C_ADDRESS*2, 0, 0, data, data_size);
    delay_us(command_description.execution_time_us);

    return status;
}


/******************************************************************************/
/*!
 * Display text on LCD.
 *
 * \param       device I2C device.
 * \param       text  text to be displayed.
 * \return      result of the function
 */
static ER lcd_print_text(I2C_Handle_t * device, char * text)
{
    ER status = E_OK;
	int i;

    for(i=0; i<strlen(text); i++)
    {
        status = i2c_memwrite(device, LCD_I2C_ADDRESS*2, 0, 0, &text[i], 1);
        if(status != E_OK)
        {
            break;
        }
        delay_us(LCD_PRINT_DELAY_US);
    }

    return status;
}

/******************************************************************************/
/*!
 * Demonstrate using I2C to drive the LCD display.
 *
 */
static ER lcd_demo(I2C_Handle_t * device)
{
    ER status = E_OK;
    uint8_t param;

    // Turn display on
    if(status == E_OK)
    {
        syslog_0(LOG_NOTICE, "INFO: Turning display on.");
        status = lcd_send_command(device, LCD_COMMAND_DISPLAY_ON, NULL);
    }

    // Clear screen
    if(status == E_OK)
    {
        syslog_0(LOG_NOTICE, "INFO: Clearing screen.");
        status = lcd_send_command(device, LCD_COMMAND_CLEAR_SCREEN, NULL);
    }

    // Turn cursor on
    if(status == E_OK)
    {
        syslog_0(LOG_NOTICE, "INFO: Turning cursor on.");
        status = lcd_send_command(device, LCD_COMMAND_BLINKING_CURSOR_ON, NULL);
    }

    // Display some text
    if(status == E_OK)
    {
        syslog_0(LOG_NOTICE, "INFO: Printing text.");
        status = lcd_print_text(device, "Hello ... ");
    }

    // Display some random number
    if(status == E_OK)
    {
        char str[5];
        syslog_0(LOG_NOTICE, "INFO: Printing random number.");
        sprintf(str, "%x", (int) rand() & 0xFFFF);
        status = lcd_print_text(device, str);
    }

    // Go to 2nd line
    if(status == E_OK)
    {
        syslog_0(LOG_NOTICE, "INFO: Moving cursor to the 2nd line.");
        param = LCD_POS_2ND_LINE;
        status = lcd_send_command(device, LCD_COMMAND_SET_CURSOR, &param);
    }

    // Display some text
    if(status == E_OK)
    {
        syslog_0(LOG_NOTICE, "INFO: Printing text.\n");
        status = lcd_print_text(device, "good bye!");
    }

    return status;
}

/******************************************************************************/
/*!
 * LCD demo wrapper.
 *
 * \return      result of the function
 */
ER i2c_demo_master_lcd(void)
{
    ER status = E_OK;
    I2C_Handle_t *device;
    I2C_Init_t cfg;

    syslog_0(LOG_NOTICE, "INFO: I2C Demo Master LCD started.");

    cfg.AddressingMode  = I2C_ADDRESSINGMODE_7BIT;
	cfg.ClockSpeed      = LCD_I2C_SPEED;
	cfg.OwnAddress1     = 0;
	cfg.GeneralCallMode = 0;
	cfg.SdaHoldClock    = 8;
    cfg.restart_enable  = I2C_RESATRT_ENABLE;
    cfg.fs_spklen       = ALT_I2C_SS_DEFAULT_SPKLEN;
	cfg.semid     = I2CTRS1_SEM;
	cfg.semlock   = I2CLOC1_SEM;

    // Init I2C module
	device = i2c_init(I2C1_PORTID, &cfg);
	if(device == NULL)
		status = E_PAR;
	else{
		device->writecallback = HAL_I2C_MemTxCpltCallback;
		device->readcallback  = HAL_I2C_MemRxCpltCallback;
		device->errorcallback = HAL_I2C_ErrorCallback;
	}
	dly_tsk(200);

    // Demonstrate LCD operation
    if(status == E_OK)
    {
        status = lcd_demo(device);
    }

    // I2C Cleanup
    if(status == E_OK)
    {
        i2c_deinit(device);
    }

    if(status == E_OK)
    {
        syslog_0(LOG_NOTICE, "INFO: I2C Demo Master LCD succeeded.\n");
    }
    else
    {
        syslog_0(LOG_NOTICE, "INFO: I2C Demo Master LCD failed.\n");
    }

	return status;
}
