/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2015 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2017 by TOPPERS PROJECT Educational Working Group.
 * 
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  $Id: sdcard_check.c 2416 2016-01-30 11:34:32Z ertl-hiro $
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include "storagedevice.h"
#include "device.h"
#include "sdmmc.h"
#include "stdio.h"
#include "ff.h"
#include "fcntl.h"
#include "sdcard_check.h"

/*
 *  サービスコールのエラーのログ出力
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))


#define ROOT_DIR	"0:"
#define FILENAME	ROOT_DIR "/data.txt"
#define STAT_FILENAME	ROOT_DIR "/stat.txt"

#define SUB_DIR1	ROOT_DIR "/sub1"
#define SUB_DIR2	ROOT_DIR "/sub1/sub2"
#define SUB_DIR3	ROOT_DIR "/sub1/sub2/sub3"

#define SECTOR_SIZE	512UL	/*  セクタ・サイズ（バイト単位）  */
#define WAIT_SYSLOG	500	/*  syslog()の表示待ち時間  */
#define WORD_MAX	0xffffUL	/*  WORD型の最大値  */
#define BUF_SIZE	300

/* sys/dirent.h */
struct dirent {
  unsigned long  d_fileno;		/* file number of entry */
  unsigned short d_reclen;		/* length of this record */
  unsigned char  d_type; 		/* file type, see below */
  unsigned char  d_namlen;		/* length of string in d_name */
#define	MAXNAMLEN	255
  char	d_name[MAXNAMLEN + 1];	/* name must be no longer than this */
};

/* File types */
#define DT_UNKMOWN  0
#define DT_FIFO     1
#define DT_CHAR     2
#define DT_DIR      4
#define DT_BLK      6
#define DT_REG      8
#define DT_LNK      10
#define DT_SOCK     12
#define DT_WHT      14


#define SD_OK   E_OK

const static char test_data[] = 
  "0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef"
  "0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef"
  "0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef"
  "0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef";

static char buff[BUF_SIZE + 1];
static int  test_mode;


static void storagedev_notice(StorageDevice_t *psdev, bool_t sw)
{
	syslog_2(LOG_NOTICE, "SD-CARD[%08x] on/off(%d) !", psdev, sw);
}

static int get_fat_time(void)
{
	int    reqtime = 0;

	reqtime = 0x00210000;
	return reqtime;
}

/*
 *  書き込みと読み出しテスト
 *　　複数クラスタにまたがるようにする。
 *　　
 *　　引数
 *　　　FATFS *fs：ファイルシステムのワークエリアの先頭アドレス
 */
static 
int test_read_write(int clstsize)
{
	int	cmp;
	uint32_t	times, i;
	uint32_t	n;
	size_t	data_length, total;
	FILE    *fid;
	int     result;

	syslog_0(LOG_NOTICE, "[file read/write test]");
	fid = fopen(FILENAME, "rwb");
	assert(fid != NULL);

	/*
	 *　書き込むデータサイズの計算
	 *　　・複数クラスタにまたがるようにする。
	 *　　・上記のopen()呼び出しで、ワークエリアが初期化済みで
	 *　　　あることが保証される。
	 */
	data_length = strlen(test_data);
	assert(data_length <= WORD_MAX);
	total = clstsize * SECTOR_SIZE;
	assert((total / data_length) <= (INT_MAX - 2UL));
	times = (uint32_t)((total / data_length) + 2UL);
#if 1	/* ROI DEBUG */
	times = 40;
	syslog_1(LOG_NOTICE, "## times(%d) ##", times);
#endif	/* ROI DEBUG */

	/*  書き込みテスト  */
	for(i = 0; i < times; ++i) {
		n = fwrite(test_data, data_length, 1, fid);
		if (n != 1) {
			syslog_1(LOG_ERROR, "test write: n(%d) Error!", n);
			syslog_1(LOG_ERROR, "   %dth trial is failed.", i);
			syslog_0(LOG_NOTICE, "test read: Skip: OK");
			dly_tsk(WAIT_SYSLOG);
			fclose(fid);
			return -1;
		}
	}
	syslog_0(LOG_NOTICE, "test write: OK");
	result = fclose(fid);
	assert(result == 0);

	fid = fopen(FILENAME, "rb");
	assert(fid != NULL);

	/*  読み出しテスト  */
	for(i = 0; i < times; ++i) {
		buff[0] = '\0';
		buff[BUF_SIZE] = '\0';
		n = fread(buff, data_length, 1, fid);
		assert(n == 1);
		cmp = strncmp(test_data, buff, data_length);
		assert(cmp == 0);
	}
	syslog_0(LOG_NOTICE, "test read: OK");
	result = fclose(fid);
	assert(result == 0);
	return 0;
}

/*
 *  seekのテスト
 *　　複数クラスタにまたがるようにする。
 *　　
 *　　引数
 *　　　FATFS *fs：ファイルシステムのワークエリアの先頭アドレス
 */
static 
int test_seek(int clstsize)
{
	int	cmp;
	UINT	n;
	size_t	data_length;
	long	cluster_size;
	FILE    *fid;
	long    offset;
	int     result;

	/*
	 *　シークするオフセットの計算
	 *　　第２クラスタに到達するようにする。
	 */
	data_length = strlen(test_data);
	cluster_size = clstsize * SECTOR_SIZE;
	offset = cluster_size - (cluster_size % data_length);

	fid = fopen(FILENAME, "rb");
	assert(fid != NULL);

	result =fseek(fid, offset, SEEK_SET);
	assert(result == offset);

	buff[0] = '\0';
	buff[BUF_SIZE] = '\0';
	n = fread(buff, data_length, 1, fid);
	if (n != 1) {
		syslog_0(LOG_ERROR, "test seek: Error(data length error)");
		dly_tsk(WAIT_SYSLOG);
		fclose(fid);
		return -1;
	}
	cmp = strncmp(test_data, buff, data_length);
	assert(cmp == 0);

	syslog_0(LOG_NOTICE, "test seek: OK");
	fclose(fid);
	return 0;
}

static 
int test_mkdir(void)
{
	int result;
	void *dirent;

	result = mkdir(SUB_DIR1, 0x777);
	assert(result == 0);
	dirent = opendir(SUB_DIR1);
	assert(dirent != NULL);
	closedir(dirent);

	result = mkdir(SUB_DIR2, 0x777);
	assert(result == 0);
	dirent = opendir(SUB_DIR2);
	assert(dirent != NULL);
	closedir(dirent);

	result = mkdir(SUB_DIR3, 0x777);
	assert(result == 0);
	dirent = opendir(SUB_DIR3);
	assert(dirent != 0);
	closedir(dirent);

	syslog_0(LOG_NOTICE, "test mkdir: OK");
#if 0	/* ROI DEBUG */
	rmdir(SUB_DIR3);
	rmdir(SUB_DIR2);
	rmdir(SUB_DIR1);
#endif	/* ROI DEBUG */
	return 0;
}

/*
 *　指定されたディレクトリ以下のディレクトリとファイルを
 *　（再帰的に）すべて表示する。
 */
static 
int scan_files (char* path)
{
    size_t i;
	void *dir;
	struct dirent *pdirent;

    if ((dir = opendir(path)) != NULL) {
		syslog_1(LOG_NOTICE, "Dir: %s/", path);

        i = strlen(path);
        while ((pdirent = readdir(dir)) != NULL && pdirent->d_name[0] >= ' ') {
            if (pdirent->d_type == DT_DIR) {
                sprintf(path+i, "/%s", pdirent->d_name);
                scan_files(path);
                *(path+i) = '\0';
            } else {
				syslog_2(LOG_NOTICE, "File: %s/%s", path, pdirent->d_name);
            }
        }
		closedir(dir);
    }
	return 0;
}

static 
int test_dir(void)
{
	strcpy(buff, ROOT_DIR);
	scan_files(buff);
	rmdir(SUB_DIR3);
	rmdir(SUB_DIR2);
	rmdir(SUB_DIR1);
	return 0;
}

/*
 *　テスト用にファイルを作る。
 */
static 
void make_file(char *filename)
{
	size_t	data_length;
	UINT	n;
	FILE    *fid;
	int     result;

	data_length = strlen(test_data);

	fid = fopen(filename, "wb");
	assert(fid != NULL);

	n = fwrite(test_data, data_length, 1, fid);
	assert(n == 1);

	result = fclose(fid);
	assert(result == 0);
}

int test_stat(void)
{
	struct stat finfo;
	int    result;
	FILE   *fid;

	make_file(STAT_FILENAME);
	/*  ファイルができていることの確認  */
	fid = fopen(STAT_FILENAME, "rb");
	assert(fid != NULL);
	fclose(fid);
	result = stat(STAT_FILENAME, &finfo);
	assert(result == 0);

	// Set read-only flag , clear archive flag and others are not changed.
	result = chmod(STAT_FILENAME, 0x222);
	assert(result == 0);

	result = stat(STAT_FILENAME, &finfo);
	assert(result == 0);

	syslog_1(LOG_NOTICE, "Check read only attribute[%s].", STAT_FILENAME);
	syslog_4(LOG_NOTICE, "fsize=%d fdate=0x%x ftime=0x%x attr=0x%x", finfo.st_size, finfo.st_atime, finfo.st_mtime, finfo.st_mode);
	syslog_0(LOG_NOTICE, "test stat: OK");
	unlink(STAT_FILENAME);
	return 0;
}


void sw_handle(void)
{
	syslog_0(LOG_NOTICE, "## sw int ##");
	test_mode++;
}

/*
 *  メインタスク
 */
void main_task(intptr_t exinf)
{
	StorageDevice_t *psdev;
	FATFS   *fs;
	ER ercd;
	int i;

	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG)));
	syslog(LOG_NOTICE, "Sample program starts (exinf = %d).", (int_t) exinf);

	/*
	 *  シリアルポートの初期化
	 *
	 *  システムログタスクと同じシリアルポートを使う場合など，シリアル
	 *  ポートがオープン済みの場合にはここでE_OBJエラーになるが，支障は
	 *  ない．
	 */
	ercd = serial_opn_por(TASK_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(TASK_PORTID,
							(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));

	psdev = SDMGetStorageDevice(0);
	psdev->_sdev_notice = storagedev_notice;
	SDeviceHead._get_datetime = get_fat_time;

	test_mode = 0;
	for(i = 0 ; i < 500 ; i++){
		dly_tsk(2000);
		syslog_0(LOG_NOTICE, "SDCARD TEST: PUSH USER SW !");
		while(test_mode == 0 || psdev->_sdev_local[0] == 0){
			dly_tsk(100);
		}

		fs = (FATFS *)psdev->_sdev_local[0];
		syslog_1(LOG_NOTICE, "## fs[%08x] ##", fs);
		test_read_write(fs->csize);
		test_seek(fs->csize);
		test_mkdir();
		test_dir();
		test_stat();
		syslog_0(LOG_NOTICE, "TEST END");
		dly_tsk(500);
		test_mode = 0;
	}

	syslog(LOG_NOTICE, "Sample program ends.");
	SVC_PERROR(ext_ker());
	assert(0);
}
