/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2017-2019 by TOPPERS PROJECT Educational Working Group.
 * 
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  @(#) $Id: fpgai2c.c 698 2019-04-18 17:45:39Z roi $
 */
/*
 *
 *  FPGA-I2Cドライバ関数群
 *
 */
#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sil.h>
#include <target_syssvc.h>
#include "fpgai2c.h"

#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

/*
 *  SPIOポートIDから管理ブロックを取り出すためのマクロ
 */
#define INDEX_I2C(i2cid)        ((uint_t)((i2cid) - 1))

#define SCL_DELAY               sil_dly_nse
#define BIT_WIDTH               8
#ifndef DEFAULT_FI2C_TIMEOUT
#define DEFAULT_FI2C_TIMEOUT    5000
#endif

static FPGAI2C_Handle_t I2CHandle[NUM_FPGAI2CPORT];

#define i2c_start(h)            i2c_trans_bit((h), (1), (0), (0))
#define i2c_stop(h)             i2c_trans_bit((h), (0), (1), (1))

int aaaa;

static ER
i2c_change_direction(FPGAI2C_Handle_t *hi2c, uint8_t oe)
{
	uint32_t timeout;

	if(oe == (hi2c->flags & I2C_FLAGS_OE))
		return E_OK;
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SDA_DIR_REG), oe);
	timeout = hi2c->timeout * 10000;
	while((sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SDA_DIR_REG)) & 1) != oe){
		sil_dly_nse(100);
		if((--timeout) == 0){
			syslog_1(LOG_ERROR, "i2c_change_direction timeout(%d) !", oe);
			return E_TMOUT;
		}
	}
	hi2c->flags = (hi2c->flags & ~I2C_FLAGS_OE) | oe;
	return E_OK;
}

static ER
i2c_trans_bit(FPGAI2C_Handle_t *hi2c, uint8_t data1, uint8_t data2, uint8_t clk)
{
	ER  ercd;
	uint32_t tmp;

	SCL_DELAY(hi2c->clkCount);
	ercd = i2c_change_direction(hi2c, 1);
	if(ercd != E_OK)
		return ercd;

	/* SDAT=D1送信 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SDA_DATA_REG), (data1 & 1));
	tmp = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SDA_DATA_REG));
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SCL_DATA_REG), 1);
	tmp = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SCL_DATA_REG));
	SCL_DELAY(hi2c->clkCount);
	/* SDAT=D2..SCLK=0 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SDA_DATA_REG), (data2 & 1));
	tmp = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SDA_DATA_REG));
	SCL_DELAY(hi2c->clkCount);
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SCL_DATA_REG), (clk & 1));
	tmp = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SCL_DATA_REG));
	SCL_DELAY(hi2c->clkCount);
	(void)(tmp);
	return E_OK;
}

static ER
i2c_send_bit(FPGAI2C_Handle_t *hi2c, uint8_t data)
{
	ER  ercd;
	uint32_t tmp;
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SCL_DATA_REG), 0);
	tmp = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SCL_DATA_REG));
	ercd = i2c_change_direction(hi2c, 1);
	if(ercd != E_OK)
		return ercd;
	/* SDAT=送信 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SDA_DATA_REG), (data & 1));
	tmp = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SDA_DATA_REG));
	SCL_DELAY(hi2c->clkCount);

	/* SDAT= 0..SCLK=0 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SCL_DATA_REG), 1);
	tmp = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SCL_DATA_REG));
	SCL_DELAY(hi2c->clkCount * 2);
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SCL_DATA_REG), 0);
	tmp = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SCL_DATA_REG));
	SCL_DELAY(hi2c->clkCount);
	(void)(tmp);
	return E_OK;
}

static int
i2c_recv_bit(FPGAI2C_Handle_t *hi2c)
{
	ER  ercd;
	int data;
	uint32_t tmp;
	/* SCLK= 0 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SCL_DATA_REG), 0);
	tmp = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SCL_DATA_REG));
	SCL_DELAY(hi2c->clkCount);
	/* SCLK=1 */
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SCL_DATA_REG), 1);
	tmp = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SDA_DATA_REG));
	/* SDAT入力 */
	ercd = i2c_change_direction(hi2c, 0);
	if(ercd != E_OK)
		return ercd;
	SCL_DELAY(hi2c->clkCount);
	/* SDAT=受信 */
	data = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SDA_DATA_REG)) & 1;
	SCL_DELAY(hi2c->clkCount);
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SCL_DATA_REG), 0);
	tmp = sil_rew_mem((uint32_t *)(hi2c->base+TOFF_I2C_SCL_DATA_REG));
	(void)(tmp);
	return data;
}

static int
i2c_send_byte(FPGAI2C_Handle_t *hi2c, uint8_t data)
{
	int     i;

	for(i = 0 ; i < BIT_WIDTH ; i++){
		i2c_send_bit(hi2c, (data >> (7-i)) & 1);
	}
	return i2c_recv_bit(hi2c);
}

static uint8_t
i2c_recv_byte(FPGAI2C_Handle_t *hi2c)
{
	uint8_t data = 0;
	int     i;

	for(i = 0 ; i < BIT_WIDTH ; i++){
		data <<= 1;
		data |= i2c_recv_bit(hi2c) & 1;
	}
	i2c_send_bit(hi2c, 1);
	return data;
}


/*
 *  FPGAI2C初期設定
 *  parameter1  port: FPGASPIポート番号
 *  return I2Cハンドラへのポインタ、NULLでエラー
 */
FPGAI2C_Handle_t *
fpgai2c_init(ID portid, uint32_t speed)

{
	FPGAI2C_Handle_t * hi2c;
	uint32_t no;

	if(portid > NUM_FPGAI2CPORT || portid == 0)
        return NULL;
	no = INDEX_I2C(portid);
	hi2c = &I2CHandle[no];
	if(hi2c->base != 0 || speed < 1000)
		return NULL;

	memset(hi2c, 0, sizeof(FPGAI2C_Handle_t));
    hi2c->base = TADR_FPGAI2C_BASE;
	hi2c->portid = portid;
	hi2c->clkCount = 250*1000*1000 / speed;
	hi2c->flags  = 0;
	hi2c->timeout = DEFAULT_FI2C_TIMEOUT;
	i2c_change_direction(hi2c, 1);
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SCL_DATA_REG), 1);
	sil_wrw_mem((uint32_t *)(hi2c->base+TOFF_I2C_SDA_DATA_REG), 1);
	return hi2c;
}

/*
 *  FPGAI2C終了設定
 *  parameter1  hi2c: I2Cハンドラへのポインタ
 *  return ERコード
 */
ER
fpgai2c_deinit(FPGAI2C_Handle_t * hi2c)
{
	if(hi2c == NULL)
		return E_PAR;
	if(hi2c->portid == 0 || hi2c->portid > NUM_FPGAI2CPORT)
		return E_PAR;
	hi2c->base   = 0;
    return E_OK;
}


/*
 *  FPGAI2Cデータリード
 *  parameter1 hi2c:       I2Cハンドラへのポインタ
 *  parameter2 DevAddress: スレーブアドレス
 *  parameter3 MemAddress: メモリアドレス
 *  parameter4 MemAddSize: メモリアドレスサイズ
 *  parameter5 pData:      読み出しバッファへのポインタ
 *  parameter6 Size:       読み出しサイズ
 *  return                 正常終了ならE_OK
 */
ER
fpgai2c_memread(FPGAI2C_Handle_t *hi2c, uint16_t devaddr, uint16_t addr, uint16_t addrsize, void * data, uint32_t Size)
{
	uint8_t *bdata = data;
	int     i;

	syslog_5(LOG_DEBUG, "fpgai2c_memread(%02x, %04x, %d, %08x, %d)", devaddr, addr, addrsize, data, Size);
	if(data == NULL || Size == 0 || hi2c == NULL)
		return E_PAR;

	/*
	 *  書き込みロック
	 */
	if(hi2c->semlock != 0)
		wai_sem(hi2c->semlock);

	/*
	 *  スレーブアドレス設定
	 */
	i2c_start(hi2c);
	if(i2c_send_byte(hi2c, devaddr & ~1) == 0){
		i2c_stop(hi2c);
		return E_OBJ;
	}

	/*
	 *  アドレス設定があれば送信
	 */
	if(addrsize == 1){
		if(i2c_send_byte(hi2c, addr) == 0){
			i2c_stop(hi2c);
			return E_OBJ;
		}
	}
	else if(addrsize > 1){
		if(i2c_send_byte(hi2c, addr>>8) == 0){
			i2c_stop(hi2c);
			return E_OBJ;
		}
		if(i2c_send_byte(hi2c, addr & 0xff) == 0){
			i2c_stop(hi2c);
			return E_OBJ;
		}
	}

	/*
	 *  受信設定と受信終了待ち
	 */
	if(i2c_send_byte(hi2c, devaddr | 1) == 0){
		i2c_stop(hi2c);
		return E_OBJ;
	}
	for(i = 0 ; i < Size; i++){
		*bdata++ = i2c_recv_byte(hi2c);
	}
	i2c_stop(hi2c);

	/*
	 *  書き込みロック解除
	 */
	if(hi2c->semlock != 0)
		sig_sem(hi2c->semlock);
	return E_OK;
}

/*
 *  FPGAI2Cデータライト
 *  parameter1 hi2c:       I2Cハンドラへのポインタ
 *  parameter2 DevAddress: スレーブアドレス
 *  parameter3 MemAddress: メモリアドレス
 *  parameter4 MemAddSize: メモリアドレスサイズ
 *  parameter5 pData:      書込みバッファへのポインタ
 *  parameter6 Size:       書込みサイズ
 *  return                 正常終了ならE_OK
 */
ER
fpgai2c_memwrite(FPGAI2C_Handle_t *hi2c, uint16_t devaddr, uint16_t addr, uint16_t addrsize, const void * data, uint32_t Size)
{
	const uint8_t *bdata = data;
	ER    ercd = E_OK;
	int   i;

	if(data == NULL || Size == 0 || hi2c == NULL)
		return E_PAR;

	/*
	 *  書き込みロック
	 */
	if(hi2c->semlock != 0)
		wai_sem(hi2c->semlock);

	/*
	 *  スレーブアドレス設定
	 */
	i2c_start(hi2c);
	if(i2c_send_byte(hi2c, devaddr & ~1) == 0){
		i2c_stop(hi2c);
		return E_OBJ;
	}

	/*
	 *  アドレス設定があれば送信
	 */
	if(addrsize == 1){
		if(i2c_send_byte(hi2c, addr) == 0){
			i2c_stop(hi2c);
			return E_OBJ;
		}
	}
	else if(addrsize > 1){
		if(i2c_send_byte(hi2c, addr>>8) == 0){
			i2c_stop(hi2c);
			return E_OBJ;
		}
		if(i2c_send_byte(hi2c, addr & 0xff) == 0){
			i2c_stop(hi2c);
			return E_OBJ;
		}
	}

	/*
	 *  送信設定と送信終了待ち
	 */
	for(i = 0 ; i < Size ; i++){
		if(i2c_send_byte(hi2c, *bdata++) == 0){
			ercd = E_OBJ;
			break;
		}
	}
	i2c_stop(hi2c);

	/*
	 *  書き込みロック解除
	 */
	if(hi2c->semlock != 0)
		sig_sem(hi2c->semlock);
	return ercd;
}

