/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2017 by TOPPERS PROJECT Educational Working Group.
 * 
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  @(#) $Id: fpgaspi.c 698 2017-08-21 12:24:16Z roi $
 */
/*
 *
 *  FPGA-SPIドライバ関数群
 *
 */
#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sil.h>
#include <target_syssvc.h>
#include "fpgaspi.h"

#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

/*
 *  SPIOポートIDから管理ブロックを取り出すためのマクロ
 */
#define INDEX_SPI(spiid)        ((uint_t)((spiid) - 1))

#ifndef DEFAULT_FSPI_TIMEOUT
#define DEFAULT_FSPI_TIMEOUT    5000
#endif

static FPGASPI_Handle_t SpiHandle[NUM_FPGASPIPORT];

/*
 *  FPGASPI初期設定
 *  parameter1  port: FPGASPIポート番号
 *  return SPIハンドラへのポインタ、NULLでエラー
 */
FPGASPI_Handle_t *
fpgaspi_init(ID portid)

{
	FPGASPI_Handle_t * hspi;
	uint32_t no;

	if(portid > NUM_FPGASPIPORT || portid == 0)
        return NULL;
	no = INDEX_SPI(portid);
	hspi = &SpiHandle[no];
	if(hspi->base != 0)
		return NULL;

	memset(hspi, 0, sizeof(FPGASPI_Handle_t));
    hspi->base = TADR_FPGASPI_BASE;
	hspi->portid = portid;
	hspi->timeout = DEFAULT_FSPI_TIMEOUT;
	return hspi;
}

/*
 *  SPI終了設定
 *  parameter1  hspi: SPIハンドラへのポインタ
 *  return ERコード
 */
ER
fpgaspi_deinit(FPGASPI_Handle_t * hspi)
{
	if(hspi == NULL)
		return E_PAR;
	if(hspi->portid == 0 || hspi->portid > NUM_FPGASPIPORT)
		return E_PAR;
	hspi->base   = 0;
    return E_OK;
}


/*
 *  SPI送受信実行関数
 *  parameter1  hspi: SPIハンドラへのポインタ
 *  parameter2  slave_select: スレーブ指定
 *  parameter3  ptxdata: 送信バッファへのポインタ(DMAなら32バイトアライン)
 *  parameter4  prxdata: 受信バッファへのポインタ(DMAなら32バイトアライン)
 *  parameter5  length: 送信サイズ
 *  return ERコード
 */
ER
fpgaspi_transrecv(FPGASPI_Handle_t *hspi, uint32_t slave, uint8_t *tx_buf, uint8_t *rx_buf, uint32_t len)
{
	ER ercd = E_OK;
	uint32_t timeout = 0;
	uint32_t tmp;

	if(hspi == NULL)
		return E_PAR;
	if(tx_buf == NULL && rx_buf == NULL)
		return E_PAR;
	if(len == 0)
		return E_OK;
	if(hspi->semlock != 0)
		wai_sem(hspi->semlock);

	/*
	 *  スレーブビットをセット
	 */
	sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_SLAVE_SEL_REG), 1 << slave);

	/*
	 *  トグル要求がなければ、SSOビットをセット
	 */
	if((hspi->flags & AVALON_SPI_COMMAND_TOGGLE_SS_N) == 0)
		sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_CONTROL_REG), AVALON_SPI_CONTROL_SSO);

	/*
	 *  受信データ・ダミー読み込み
	 */
	tmp = sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPI_RXDATA_REG));
	(void)(tmp);

	hspi->txBuff  = tx_buf;
	hspi->rxBuff  = rx_buf;
	hspi->size    = len;
	hspi->txCount = 0;
	hspi->rxCount = 0;
	sil_orw_mem((uint32_t *)(hspi->base+TOFF_SPI_CONTROL_REG), (AVALON_SPI_CONTROL_ITRDY | AVALON_SPI_CONTROL_IRRDY));

	while(hspi->rxCount < hspi->size){
		if(hspi->semid != 0)
			twai_sem(hspi->semid, 1);
		else
			dly_tsk(1);
		if(timeout++ > hspi->timeout){
			ercd = E_TMOUT;
			break;
		}
	}

	/*
	 *  マージモードでなければ、SSOをクリア
	 */
	if((hspi->flags & AVALON_SPI_COMMAND_MERGE) == 0)
		sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPI_CONTROL_REG), AVALON_SPI_CONTROL_SSO);
	if(hspi->semlock != 0)
		sig_sem(hspi->semlock);
	return ercd;
}

/*
 *  SPI割込みサービスルーチン(マスタースレーブ共通)
 */
void
spim_fpga_isr(intptr_t exinf)
{
	FPGASPI_Handle_t *hspi = &SpiHandle[INDEX_SPI(exinf)];
	uint32_t status, tmp;

	status = sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPI_STATUS_REG));
	if((status & AVALON_SPI_STATUS_TRDY) != 0){
		if(hspi->txCount < hspi->size){
			if(hspi->txBuff != NULL)
				sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_TXDATA_REG), *hspi->txBuff++);
			else
				sil_wrw_mem((uint32_t *)(hspi->base+TOFF_SPI_TXDATA_REG), 0x000000FF);
			hspi->txCount++;
		}
		else
			sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPI_CONTROL_REG), AVALON_SPI_CONTROL_ITRDY);
	}
	if((status & AVALON_SPI_STATUS_RRDY) != 0){
		tmp = sil_rew_mem((uint32_t *)(hspi->base+TOFF_SPI_RXDATA_REG));
		if(hspi->rxCount < hspi->size){
			if(hspi->rxBuff != NULL)
				*hspi->rxBuff++ = (uint8_t)tmp;
			hspi->rxCount++;
		}
		if(hspi->rxCount >= hspi->size)
			sil_andw_mem((uint32_t *)(hspi->base+TOFF_SPI_CONTROL_REG), AVALON_SPI_CONTROL_IRRDY);
	}
	if((status & AVALON_SPI_STATUS_TMT) != 0){
		if(hspi->rxCount >= hspi->size){
			if(hspi->semid != 0)
				isig_sem(hspi->semid);
		}
	}
}

