/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2015 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2017-2020 by TOPPERS PROJECT Educational Working Group.
 * 
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  @(#) $Id: fpgauart.c 64 2020-04-15 10:56:10Z roi $
 */

/*
 *  FPGAUARTドライバ関数群
 */
#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <sil.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include <target_syssvc.h>
#include "fpgauart.h"

#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

/*
 *  UARTポートIDから管理ブロックを取り出すためのマクロ
 */
#define INDEX_UART(uartid)      ((uint_t)((uartid) - 1))

#ifndef QUEUE_BUFSIZE
#define QUEUE_BUFSIZE  1024
#endif

#ifndef CTS1_PINNO
#define CTS1_PINNO     2	/* D2 */
#endif
#ifndef RTS1_PINNO
#define RTS1_PINNO     3	/* D3 */
#endif

#define LIMIT1_SIZE    ((QUEUE_BUFSIZE*7)/8)
#define LIMIT2_SIZE    ((QUEUE_BUFSIZE*6)/8)

static queue_t txque[NUM_FPGAUARTPORT];
static queue_t rxque[NUM_FPGAUARTPORT];
static uint8_t txbuffer[NUM_FPGAUARTPORT][QUEUE_BUFSIZE];
static uint8_t rxbuffer[NUM_FPGAUARTPORT][QUEUE_BUFSIZE];

static const FPGAUART_PortControlBlock uart_pcb_table[NUM_FPGAUARTPORT] = {
	{TADR_FPGAUART_BASE, INTNO_F2S_FPGA_IRQ0, CTS1_PINNO, RTS1_PINNO}
};

static FPGAUART_Handle_t Huart[NUM_FPGAUARTPORT];


static void lmemcpy(volatile void *d, const volatile void *s, int len)
{
  volatile char *src = (volatile char *)s;
  volatile char *dst = d;

  while(len > 0){
    *dst++ = *src++;
    len--;
  }
}

/*
 *  送信割込み許可
 */
static void
fpgauart_enable_send(FPGAUART_Handle_t *huart)
{
	if((huart->ctrl & AVALON_UART_CONTROL_TRDY) != 0)
		return;
	huart->ctrl |= AVALON_UART_CONTROL_TRDY;
	sil_wrw_mem((uint32_t *)(huart->base+TOFF_FPGAUART_CONTROL), huart->ctrl);
}

/*
 *  送信割込み禁止
 */
static void
fpgauart_disable_send(FPGAUART_Handle_t *huart)
{
	huart->ctrl &= ~AVALON_UART_CONTROL_TRDY;
	sil_wrw_mem((uint32_t *)(huart->base+TOFF_FPGAUART_CONTROL), huart->ctrl);
}

/*
 *  RTS受信許可
 */
static void
fpgauart_enable_rts(FPGAUART_Handle_t *huart)
{
	if(huart->pcb->rtspin == 0)
		return;
	digitalWrite(huart->pcb->rtspin, 0);
}

/*
 *  RTS受信不可
 */
static void
fpgauart_disable_rts(FPGAUART_Handle_t *huart)
{
	if(huart->pcb->rtspin == 0)
		return;
	digitalWrite(huart->pcb->rtspin, 1);
}

/*
 *  CTSチェック
 */
static bool_t
fpgauart_sens_cts(FPGAUART_Handle_t *huart)
{
	if(huart->pcb->ctspin == 0)
		return true;
	return (bool_t)digitalRead(huart->pcb->ctspin);
}

/*
 *  拡張UART割込みハンドラ
 */
void
fpgauart_isr(intptr_t exinf)
{
	FPGAUART_Handle_t *huart = &Huart[INDEX_UART(exinf)];
	uint32_t status, tmp;
	queue_t  *que;
	int      bsize;

	status = sil_rew_mem((uint32_t *)(huart->base+TOFF_FPGAUART_STATUS));
	sil_wrw_mem((uint32_t *)(huart->base+TOFF_FPGAUART_STATUS), 0);
	tmp = sil_rew_mem((uint32_t *)(huart->base+TOFF_FPGAUART_STATUS));
	if((status & (AVALON_UART_STATUS_PE | AVALON_UART_STATUS_FE)) != 0)
		return;
	if((status & AVALON_UART_STATUS_RRDY) != 0){
		que = huart->rque;
		bsize = que->top - que->tail;
		if(bsize < 0)
			bsize += que->size;
		if(bsize < (que->size-1)){
			tmp = sil_rew_mem((uint32_t *)(huart->base+TOFF_FPGAUART_RXDATA));
			que->buffer[que->top++] = (uint8_t)tmp;
			if(que->top >= que->size)
				que->top = 0;
			if(huart->rxcallback != NULL)
				huart->rxcallback(huart);
		}
		if(bsize >= LIMIT1_SIZE)
			fpgauart_disable_rts(huart);
	}
	if((status & AVALON_UART_STATUS_TRDY) != 0){
		que = huart->tque;
		if(fpgauart_sens_cts(huart) && que->top != que->tail){
			sil_wrw_mem((uint32_t *)(huart->base+TOFF_FPGAUART_TXDATA), que->buffer[que->tail++]);
			if(que->tail >= que->size)
				que->tail -= que->size;
			if(que->top == que->tail)
				fpgauart_disable_send(huart);
		}
		else
			fpgauart_disable_send(huart);
	}
}

/*
 *  拡張UART初期化
 */
FPGAUART_Handle_t *
fpgauart_init(ID portid)
{
	const FPGAUART_PortControlBlock *pcb;
	FPGAUART_Handle_t *huart;
	uint32_t no = INDEX_UART(portid);

	if(portid < 0 || portid > NUM_FPGAUARTPORT)
		return NULL;

	txque[no].size = QUEUE_BUFSIZE;
	txque[no].top  = 0;
	txque[no].tail = 0;
	txque[no].buffer = txbuffer[no];
	rxque[no].size = QUEUE_BUFSIZE;
	rxque[no].top  = 0;
	rxque[no].tail = 0;
	rxque[no].buffer = rxbuffer[no];

	pcb   = &uart_pcb_table[no];
	huart = &Huart[no];
	huart->pcb  = pcb;
	huart->base = pcb->base;
	huart->tque = &txque[no];
	huart->rque = &rxque[no];

	if(pcb->rtspin != 0)
		pinMode(pcb->rtspin, OUTPUT);
	if(pcb->ctspin != 0)
		pinMode(pcb->ctspin, INPUT);

	huart->ctrl = AVALON_UART_CONTROL_RTS |
				AVALON_UART_CONTROL_RRDY |
				AVALON_UART_CONTROL_DCTS;
	sil_wrw_mem((uint32_t *)(huart->base+TOFF_FPGAUART_CONTROL), huart->ctrl);
	fpgauart_enable_rts(huart);
	return huart;
}

/*
 *  拡張UART無効化
 */
ER
fpgauart_deinit(FPGAUART_Handle_t *huart)
{
	if(huart == NULL)
		return E_PAR;
	huart->ctrl = 0;
	sil_wrw_mem((uint32_t *)(huart->base+TOFF_FPGAUART_CONTROL), huart->ctrl);
	return E_OK;
}

/*
 *  拡張UART受信
 */
int
fpgauart_recev(FPGAUART_Handle_t *huart, unsigned char *buf, int len)
{
	queue_t  *que;
	int      bsize, rlen = 0;

	if(huart == NULL)
		return 0;
	que = huart->rque;
	while(len > rlen){
		bsize = que->top - que->tail;
		if(bsize == 0)
			break;
		else if(bsize < 0)
			bsize = que->size - que->tail;
		if(bsize > (len - rlen))
			bsize = len - rlen;
		lmemcpy(buf, &que->buffer[que->tail], bsize);
		buf += bsize;
		que->tail += bsize;
		if(que->tail >= que->size)
			que->tail -= que->size;
		rlen += bsize;
	}
	return rlen;
}

/*
 *  拡張UART受信サイズ取得
 */
int
fpgauart_recsize(FPGAUART_Handle_t *huart)
{
	queue_t  *que;
	int      bsize;

	if(huart == NULL)
		return 0;
	que = huart->rque;
	bsize = que->top - que->tail;
	if(bsize < 0)
		bsize += que->size;
	if(bsize < LIMIT2_SIZE)
		fpgauart_enable_rts(huart);
	return bsize;
}

/*
 *  拡張UART送信
 */
int
fpgauart_send(FPGAUART_Handle_t *huart, unsigned char *buf, int len)
{
	queue_t  *que;
	int      bsize, i;

	if(huart == NULL)
		return 0;
	que = (queue_t *)huart->tque;
	for(i = 0 ; i < len ; i++){
		bsize = que->top - que->tail;
		if(bsize < 0)
			bsize += que->size;
		if(bsize < (que->size-1)){
			que->buffer[que->top++] = *buf++;
			if(que->top >= que->size)
				que->top -= que->size;
		}
		else{
			if(fpgauart_sens_cts(huart))
				fpgauart_enable_send(huart);
			dly_tsk(20);
		}
	}
	if(i != 0 && fpgauart_sens_cts(huart))
		fpgauart_enable_send(huart);
	return i;
}

/*
 *  CTSセンスチェック
 */
bool_t
fpgauart_cts_sens(FPGAUART_Handle_t *huart)
{
	queue_t  *que;

	if(huart == NULL)
		return false;
	que = huart->tque;
	if(que->top == que->tail || !fpgauart_sens_cts(huart))
		return false;
	fpgauart_enable_send(huart);
	return true;
}


