/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *  Copyright (C) 2015-2018 by TOPPERS PROJECT Educational Working Group.
 * 
 *  上記著作権者は，以下の (1)～(4) の条件か，Free Software Foundation 
 *  によって公表されている GNU General Public License の Version 2 に記
 *  述されている条件を満たす場合に限り，本ソフトウェア（本ソフトウェア
 *  を改変したものを含む．以下同じ）を使用・複製・改変・再配布（以下，
 *  利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，その適用可能性も
 *  含めて，いかなる保証も行わない．また，本ソフトウェアの利用により直
 *  接的または間接的に生じたいかなる損害に関しても，その責任を負わない．
 * 
 *  @(#) $Id: sdcard.c,v 1.1 2018/04/09 18:49:31 roi Exp $
 */
/*
 *  CycolneV HPS用 SDMMCドライバ
 */
#include "kernel_impl.h"
#include <t_syslog.h>
#include <t_stdlib.h>
#include <sil.h>
#include <string.h>
#include <target_syssvc.h>
#include "device.h"
#include "sdmmc.h"
#include "kernel_cfg.h"



#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

#define MIN(a, b) ((a) > (b) ? (b) : (a))

#ifndef EXT_CLK_CLKSC1_FREQCUR
#define EXT_CLK_CLKSC1_FREQCUR                 25000000
#endif
#ifndef EXT_CLK_CLKSC2_FREQCUR
#define EXT_CLK_CLKSC2_FREQCUR	               25000000
#endif
#ifndef EXT_CLK_PERIPH_FREQCUR
#define EXT_CLK_PERIPH_FREQCUR                 0
#endif

#ifndef SDMMC_TRANSFAR_SPEED
#define SDMMC_TRANSFAR_SPEED                   12500000
#endif

/*
 *  CLOCK SOURCE
 */
#define CLKMGR_PERPLL_SRC_SDMMC_E_F2S_PERIPH_REF_CLK    0x0	/* f2s_periph_ref_clk */
#define CLKMGR_PERPLL_SRC_SDMMC_E_MAIN_NAND_CLK         0x1	/* main_nand_sdmmc_clk */
#define CLKMGR_PERPLL_SRC_SDMMC_E_PERIPH_NAND_CLK       0x2	/* periph_nand_sdmmc_clk */

#define CLKMGR_PERPLL_VCO_PSRC_E_EOSC1          0x0	/* eosc1_clk */
#define CLKMGR_PERPLL_VCO_PSRC_E_EOSC2          0x1	/* eosc2_clk */
#define CLKMGR_PERPLL_VCO_PSRC_E_F2S_PERIPH_REF 0x2	/* f2s_periph_ref_clk */

/*
 *  リセットマネージャー用タイムアウト値
 */
#define SDMMC_RESET_TMO_INIT            8192

/*
 *  SDMMCモジュールリセット用タイムアウト値
 */
#define SDMMC_MAX_T_POLL_COUNT          8192
#define SDMMC_TMO_WAITER                500
#define SDMMC_CMDTIMEOUT                1000
#define SD_MAX_VOLT_TRIAL               0x0000FFFF

#define SDMMC_BLOCKSIZE                 512
#define SDMMC_DMA_SEGMENT_SIZE          SDMMC_BLOCKSIZE
#define SDMMC_DMA_DESC_COUNT            64

/*
 *  RESPONSE1エラー値
 */
#define SD_OCR_ERROR                    0xFDFFE008

/*
 *  カードタイプ定義
 */
#define SD_HIGH_CAPACITY                0x40000000
#define SD_STD_CAPACITY                 0x00000000

/*
 *  割込み設定
 */
#define SDMMC_INT_STATUS_ALL            0x0001FFFF
#define SDMMC_DMA_INT_STATUS_ALL        0x00000337

/*
 *  SD-CARDデータ長
 */
#define SCR_LENGTH                      8
#define SYSTEM_STATUS_LENGTH            64

/*
 *  FIFOのエントリィ数
 */
#define SDMMC_FIFO_NUM_ENTRIES          1024

#ifndef SDMMC_READ_INNER_DMA
#define SDMMC_DESC_COUNT                ((SDMMC_FIFO_NUM_ENTRIES*4)/SDMMC_BLOCKSIZE)
#else
#define SDMMC_DESC_COUNT                SDMMC_DMA_DESC_COUNT
#endif

/*
 *  WRITE後の待ち時間
 */
#define SDMMC_WRITE_DELAY_TIME          30	/* ms */


#define CCC_CLASS_10                    (1<<10)

#define SDMMC_CSD_SPEED_RATE_MSK        0x00000007
#define SDMMC_CSD_SPEED_TIME_MSK        0x00000078
#define SDMMC_CSD_CCC_MSK               0xFFF00000
#define SDMMC_CSD_SPEED_RATE_GET(val)   ((val & SDMMC_CSD_SPEED_RATE_MSK) >> 0)
#define SDMMC_CSD_SPEED_TIME_GET(val)   ((val & SDMMC_CSD_SPEED_TIME_MSK) >> 3)
#define SDMMC_CSD_CCC_GET(val)          ((val & SDMMC_CSD_CCC_MSK) >> 20)

#define SDMMC_SWITCH_FUNC_MODE_CHECK    (0UL << 31)

#define SDMMC_HIGH_SPEED_BUS_SPEED      1
#define SDMMC_UHS_SDR25_BUS_SPEED       1
#define SDMMC_UHS_SDR50_BUS_SPEED       2
#define SDMMC_UHS_SDR104_BUS_SPEED      3
#define SDMMC_UHS_DDR50_BUS_SPEED       4

#define SDMMC_SD_MODE_HIGH_SPEED        (1 << SDMMC_HIGH_SPEED_BUS_SPEED)
#define SDMMC_SD_MODE_UHS_SDR25         (1 << SDMMC_UHS_SDR25_BUS_SPEED)
#define SDMMC_SD_MODE_UHS_SDR50         (1 << SDMMC_UHS_SDR50_BUS_SPEED)
#define SDMMC_SD_MODE_UHS_SDR104        (1 << SDMMC_UHS_SDR104_BUS_SPEED)
#define SDMMC_SD_MODE_UHS_DDR50         (1 << SDMMC_UHS_DDR50_BUS_SPEED)

#define SDMMC_DMA_PBL_1                 0x0		/* 1 transfer unit    */
#define SDMMC_DMA_PBL_4                 0x1		/* 4 transfer units   */
#define SDMMC_DMA_PBL_8                 0x2		/* 8 transfer units   */
#define SDMMC_DMA_PBL_16                0x3     /* 16 transfer units  */
#define SDMMC_DMA_PBL_32                0x4		/* 32 transfer units  */
#define SDMMC_DMA_PBL_64                0x5		/* 64 transfer units  */
#define SDMMC_DMA_PBL_128               0x6		/* 128 transfer units */
#define SDMMC_DMA_PBL_256               0x7		/* 256 transfer units */


#ifdef SDMMC_READ_INNER_DMA
static SDMMC_DMA_BUF_DESC_t  dma_descriptors[SDMMC_DMA_DESC_COUNT] __attribute__ ((aligned (CACHE_LINE_SIZE)));
#endif
static uint8_t               sbuffer[128] __attribute__ ((aligned (CACHE_LINE_SIZE)));

static SDMMC_Handle_t SDHandle;

#define ALT_SDMMC_DMA_BUF_DESC_CACHE_SIZE (((SDMMC_DMA_DESC_COUNT*sizeof(SDMMC_DMA_BUF_DESC_t)) + CACHE_LINE_SIZE - 1) & ~(CACHE_LINE_SIZE-1))

static ER sdmmc_sendcommand(SDMMC_Handle_t *hsd, uint32_t cmd, uint32_t arg, ER (*func)(SDMMC_Handle_t *, uint32_t));
static ER sdmmc_checkrep1(SDMMC_Handle_t *hsd, uint32_t Cmd);
static ER sdmmc_checkrep2(SDMMC_Handle_t *hsd, uint32_t Cmd);
static ER sdmmc_checkrep3(SDMMC_Handle_t *hsd, uint32_t Cmd);
static ER sdmmc_checkrep6(SDMMC_Handle_t *hsd, uint32_t Cmd);
static ER sdmmc_cmd_wait(SDMMC_Handle_t *hsd);


/*
 *  SDMMCクロックの有効判定
 */
static bool_t
sdmmc_clk_is_enabled(void)
{
	bool_t status = sil_rew_mem((uint32_t *)(TADR_CLKMGR_PERPLL_BASE+TOFF_CLKMGR_PERPLL_EN));

	if((status & CLKMGR_PERPLL_EN_SDMMCCLK) != 0)
		return true;
	else
		return false;
}

/*
 *  SDMMCクロックの取得
 */
static
uint32_t sdmmc_clock(void)
{
    uint32_t temp  = 0;
	uint64_t temp1 = 0;
    uint64_t numer = 0;
    uint64_t denom = 1;

	temp = sil_rew_mem((uint32_t *)(TADR_CLKMGR_PERPLL_BASE+TOFF_CLKMGR_PERPLL_SRC)) & CLKMGR_PERPLL_SRC_SDMMC;
	if(temp == CLKMGR_PERPLL_SRC_SDMMC_E_F2S_PERIPH_REF_CLK)
		numer = (uint64_t)EXT_CLK_PERIPH_FREQCUR;
	else if(temp == CLKMGR_PERPLL_SRC_SDMMC_E_MAIN_NAND_CLK){
		uint32_t temp0;
		temp0 = sil_rew_mem((uint32_t *)(TADR_CLKMGR_MGR_BASE+TOFF_CLKMGR_BYPASS));
		if((temp0 & CLKMGR_BYPASS_MAINPLL) != 0)
			temp = EXT_CLK_CLKSC1_FREQCUR;
		else{
			temp0 = sil_rew_mem((uint32_t *)(TADR_CLKMGR_MPLL_BASE+TOFF_CLKMGR_MPLL_VCO));
			numer = (temp0 & CLKMGR_MPLL_VCO_NUMER) >> 3;
			denom = (temp0 & CLKMGR_MPLL_VCO_DENOM) >> 16;
			temp1 = (uint64_t)EXT_CLK_CLKSC1_FREQCUR;
			temp1 *= (numer + 1);
			temp1 /= (denom + 1);

			if(temp1 <= UINT32_MAX)
				temp = (uint32_t) temp1;
			else
				return 0;
		}

		numer = (uint64_t) temp;
		temp = (sil_rew_mem((uint32_t *)(TADR_CLKMGR_MPLL_BASE+TOFF_CLKMGR_MPLL_MAINNANDSDMMCCLK)) & CLKMGR_MPLL_MAINNANDSDMMCCLK_CNT) + 1;
		denom = (uint64_t) temp;
	}
	else if (temp == CLKMGR_PERPLL_SRC_SDMMC_E_PERIPH_NAND_CLK){
		uint64_t denom2 = 1;
		temp = sil_rew_mem((uint32_t *)(TADR_CLKMGR_MGR_BASE+TOFF_CLKMGR_BYPASS));
		if((temp & CLKMGR_BYPASS_PERPLL) != 0){
			if((temp & CLKMGR_BYPASS_PERPLLSRC) == 0)
				temp = EXT_CLK_CLKSC1_FREQCUR;
			else
				temp = EXT_CLK_PERIPH_FREQCUR;
		}
		else{
			temp = sil_rew_mem((uint32_t *)(TADR_CLKMGR_PERPLL_BASE+TOFF_CLKMGR_PERPLL_VCO));
			numer = (temp & CLKMGR_PERPLL_VCO_NUMER) >> 3;
			denom2 = (temp & CLKMGR_PERPLL_VCO_DENOM) >> 16;
			temp = (temp & CLKMGR_PERPLL_VCO_PSRC) >> 22;
			if (temp == CLKMGR_PERPLL_VCO_PSRC_E_EOSC1)
				temp1 = (uint64_t)EXT_CLK_CLKSC1_FREQCUR;
			else if (temp == CLKMGR_PERPLL_VCO_PSRC_E_EOSC2){
				temp1 = (uint64_t)EXT_CLK_CLKSC2_FREQCUR;
			}
			else if (temp == CLKMGR_PERPLL_VCO_PSRC_E_F2S_PERIPH_REF){
				temp1 = (uint64_t)EXT_CLK_PERIPH_FREQCUR;
			}
    
			if (temp1 != 0){
				temp1 *= (numer + 1);
				temp1 /= (denom2 + 1);
				if (temp1 <= UINT32_MAX)
					temp = (uint32_t) temp1;
				else
					return 0;
			}
		}

		numer = (uint64_t)temp;
		denom = (uint64_t)(sil_rew_mem((uint32_t *)(TADR_CLKMGR_PERPLL_BASE+TOFF_CLKMGR_PERPLL_PERNANDSDMMCCLK)) & CLKMGR_PERPLL_PERNANDSDMMCCLK_CNT) + 1;
	}
	else
		return 0;
	if(denom > 0){
		numer /= denom;
		if(numer <= UINT32_MAX)
			return (uint32_t)numer;
	}
    return 0;
}



#ifdef SDMMC_READ_INNER_DMA
/*
 *  DMAチェーンディスクリプタの初期化
 */
static void
sdmmc_desc_chain_init(SDMMC_Handle_t *hsd)
{
	uint32_t count;
	SDMMC_DMA_BUF_DESC_t * dma_desc = dma_descriptors;

	for(count = 0; count < SDMMC_DMA_DESC_COUNT; count++){
		dma_desc[count].des0.fld.own  = 0;
		dma_desc[count].des0.fld.ch   = 1;
		dma_desc[count].des0.fld.er   = 0;
		dma_desc[count].des1.fld.bs1  = 0;
		dma_desc[count].des2.fld.bap1 = 0;

		if(count == (SDMMC_DMA_DESC_COUNT - 1)){	/* 最後のディスクリプタは先頭に戻る */
			/*  If it is latest element set pointer to the ring head.*/
			dma_desc[count].des3.fld.bap2_or_next = (uint32_t) dma_desc;
		}
		else{	/* 次のディスクリプタを設定 */
			dma_desc[count].des3.fld.bap2_or_next = (uint32_t) (&dma_desc[count + 1]);
		}
	}
	hsd->pCurDesc = dma_desc;
	flushinvalidatedcache_by_addr((uint8_t *)dma_descriptors, ALT_SDMMC_DMA_BUF_DESC_CACHE_SIZE);
}

/*
 *  DMA処理用にチェーンディスクリプタのクリア
 */
static void
sdmmc_desc_chain_clear(SDMMC_Handle_t *hsd)
{
	uint32_t count;
	SDMMC_DMA_BUF_DESC_t * dma_desc = dma_descriptors;

	for (count = 0; count < SDMMC_DMA_DESC_COUNT; count++){
		dma_desc[count].des0.fld.own  = 0;
		dma_desc[count].des0.fld.dic  = 0;
		dma_desc[count].des0.fld.ld   = 0;
		dma_desc[count].des0.fld.fs   = 0;
		dma_desc[count].des1.fld.bs1  = 0;
		dma_desc[count].des2.fld.bap1 = 0;
	}
	hsd->pCurDesc = dma_desc;
	flushinvalidatedcache_by_addr((uint8_t *)dma_descriptors, ALT_SDMMC_DMA_BUF_DESC_CACHE_SIZE);
}

/*
 *  DMA転送セットアップ
 */
static ER
sdmmc_setup_dma_trans(SDMMC_Handle_t *hsd, uint32_t * buffer, size_t buf_len)
{
    ER ercd = E_OK;
    SDMMC_DMA_BUF_DESC_t *cur_dma_desc = hsd->pCurDesc;

    uint32_t cur_buffer = (uint32_t)buffer;
    uint32_t len_left = buf_len;

	flushinvalidatedcache_by_addr((uint8_t *)buffer, buf_len);
	while(len_left > 0){
		uint32_t idmac_status;
		/*
		 *  エラー発生チェック
		 */
		if(hsd->ErrorCode != SDMMC_ERROR_NONE){
            ercd = E_SYS;
            break;
		}

		/*
		 *  DMAディスクリプタ設定
		 */
        if (cur_dma_desc->des0.fld.own == 0){
			int set_len = len_left > SDMMC_DMA_SEGMENT_SIZE ? SDMMC_DMA_SEGMENT_SIZE : len_left;
			cur_dma_desc->des0.fld.dic = 1;
			cur_dma_desc->des0.fld.fs = (buf_len == len_left) ? 1 : 0;
			cur_dma_desc->des1.fld.bs1 = set_len;		/* dmaサイズ設定 */
			cur_dma_desc->des2.fld.bap1 = cur_buffer;	/* dmaアドレス設定 */
			cur_buffer += set_len;
			len_left -= set_len;
			cur_dma_desc->des0.fld.ld = (len_left == 0) ? 1 : 0;
			cur_dma_desc->des0.fld.own = 1;				/* use設定 */
			cur_dma_desc = (SDMMC_DMA_BUF_DESC_t *)cur_dma_desc->des3.fld.bap2_or_next;
		}
		flushinvalidatedcache_by_addr((uint8_t *)dma_descriptors, ALT_SDMMC_DMA_BUF_DESC_CACHE_SIZE);

		idmac_status = hsd->dstatus;
		hsd->dstatus = 0;

        /*
		 *  DU発生なら、転送レジューム設定
		 */
		if(idmac_status & SDMMC_IDSTS_DU){
			sil_wrw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_PLDMND), 0x1);
		}
		else if(idmac_status & SDMMC_IDSTS_AIS){	/* AIS発生 */
			ercd = E_SYS;
			break;
		}
	}
	if(ercd == E_OK){
		hsd->pBuffPtr  = NULL;
		hsd->XferCount = buf_len;
	}
    return ercd;
}

/*
 *  DMA転送スタート設定
 */
static void
sdmmc_dma_start(SDMMC_Handle_t *hsd, const uint32_t desc_skip_len, const uint32_t burst_len, const bool_t use_fixed_burst)
{
	uint32_t bmod_set_value = (burst_len<<8) | (desc_skip_len<<2) | (use_fixed_burst<<1);
    uint32_t bmod_set_mask = SDMMC_BMOD_PBL | SDMMC_BMOD_FB | SDMMC_BMOD_DSL;

    /*
	 *  ディスクリプタ設定と有効化
	 */
	sil_wrw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_DBADDR), (uint32_t)hsd->pCurDesc);
	sil_modw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_BMOD), bmod_set_mask, bmod_set_value);
}

/*
 *  DMAリセット
 */
static ER
sdmmc_dma_reset(SDMMC_Handle_t *hsd)
{
    uint32_t timeout = SDMMC_MAX_T_POLL_COUNT;

	/*
	 *  リセット要求
	 */
	sil_orw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_CTL), SDMMC_CTL_DMA_RST);

    /*
	 *  リセット終了待ち
	 */
	while((sil_rew_mem((uint32_t *)(hsd->base+TOFF_SDMMC_CTL)) & SDMMC_CTL_DMA_RST) != 0){
		timeout--;
		if(timeout == 0)
			return E_TMOUT;
	}
	return E_OK;
}

/*
 *  DMA割込み設定
 */
static void
sdmmc_dma_int_enable(SDMMC_Handle_t *hsd, const uint32_t mask)
{
	sil_orw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_IDINTEN),  mask);
}

#endif

/*
 *  DMA有効設定
 */
static void
sdmmc_dma_enable(SDMMC_Handle_t *hsd)
{
#ifdef SDMMC_READ_INNER_DMA
	if(!hsd->dma_enable){
		sil_orw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_CTL), SDMMC_CTL_USE_INTERNAL_DMAC);
		sil_orw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_BMOD), SDMMC_BMOD_DE);
		hsd->dma_enable = true;
	}
#endif
}

/*
 *  DMA無効設定
 */
static void
sdmmc_dma_disable(SDMMC_Handle_t *hsd)
{
	if(hsd->dma_enable){
		sil_andw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_CTL), SDMMC_CTL_USE_INTERNAL_DMAC);
		sil_andw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_BMOD), SDMMC_BMOD_DE);
		hsd->dma_enable = false;
	}
}

/*
 *  SD/MMCパワー状態チェック
 */
static bool_t
sdmmc_card_pwr_is_on(SDMMC_Handle_t *hsd)
{
	if((sil_rew_mem((uint32_t *)(hsd->base+TOFF_SDMMC_PWREN)) & SDMMC_PWREN_POWER_EN) != 0)
		return true;
	else
		return false;
}

/*
 *  SDMMCパワーオフ・カード
 */
static void
sdmmc_card_pwr_off(SDMMC_Handle_t *hsd)
{
	if(!sdmmc_card_pwr_is_on(hsd))
		return;

	/*
	 *  SD/MMC_enableクリア
	 */
	sil_andw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_PWREN), SDMMC_PWREN_POWER_EN);

	/*
	 *  SDMMC割込みステータスクリア
	 */
	sil_wrw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_RINTSTS), SDMMC_INT_STATUS_ALL);

	/*
	 *  RCA領域をクリア
	 */
	hsd->RCA = 0;
}

/*
 *  テスト カード・データビジィ
 */
static bool_t
sdmmc_is_busy(SDMMC_Handle_t *hsd)
{
	if((sil_rew_mem((uint32_t *)(hsd->base+TOFF_SDMMC_STAT)) & SDMMC_STAT_DATA_BUSY) != 0)
		return true;
	else
		return false;
}

/*
 *  テスト カード・アイドル
 */
static bool_t
sdmmc_is_idle(SDMMC_Handle_t *hsd)
{
	uint32_t mmc_state = sil_rew_mem((uint32_t *)(hsd->base+TOFF_SDMMC_STAT)) & SDMMC_STAT_CMD_FSM_STATES;
	uint32_t dma_state = sil_rew_mem((uint32_t *)(hsd->base+TOFF_SDMMC_IDSTS)) & SDMMC_IDSTS_FSM;

    if(mmc_state != 0 || dma_state != 0)
		return false;
	else
		return true;
}

/*
 *  クロックコンフィグレーション設定
 */
static ER
sdmmc_card_clk_div_set(SDMMC_Handle_t *hsd, const uint32_t clk_div)
{
    uint32_t low_power_mode;
    ER ercd = E_OK;

    if(sdmmc_is_busy(hsd))
		return E_SYS;

	/*
	 *  ローパワークロック状態を取得
	 */
	low_power_mode = sil_rew_mem((uint32_t *)(hsd->base+TOFF_SDMMC_CLKENA)) & SDMMC_CLKENA_CCLK_LOW_POWER;

    /*
	 *  cclk_enableビット無効化
	 */
	sil_wrw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_CLKENA), 0);
	ercd = sdmmc_sendcommand(hsd, MCI_CLK, 0x0, NULL);
	if(ercd != E_OK){
		syslog_0(LOG_ERROR, "SDMMC:alt_sdmmc_card_clk_disable failed");
		return ercd;
	}

	/*
	 *  sdmmc_clk_enableを無効化
	 */
	sil_andw_mem((uint32_t *)(TADR_CLKMGR_PERPLL_BASE+TOFF_CLKMGR_PERPLL_EN), CLKMGR_PERPLL_EN_SDMMCCLK);

	/*
	 *  drvselと samplselの設定
	 */
    if(clk_div == 0){
        /*  50 MHz*/
		sil_wrw_mem((uint32_t *)(TADR_SYSMGR_BASE+TOFF_SYSMGR_SDMMC_CTL),
									SYSMGR_SDMMC_CTL_SMPLSEL_E_DEGREES90 | SYSMGR_SDMMC_CTL_DRVSEL_E_DEGREES90);
	}
	else{
        /*  25MHz and lower*/
		sil_wrw_mem((uint32_t *)(TADR_SYSMGR_BASE+TOFF_SYSMGR_SDMMC_CTL),
									SYSMGR_SDMMC_CTL_SMPLSEL_E_DEGREES180 | SYSMGR_SDMMC_CTL_DRVSEL_E_DEGREES180);
	}

    /*
	 *  sdmmc_clk_enableを有効化
	 */
	sil_orw_mem((uint32_t *)(TADR_CLKMGR_PERPLL_BASE+TOFF_CLKMGR_PERPLL_EN), CLKMGR_PERPLL_EN_SDMMCCLK);
	sil_wrw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_CLKDIV), (clk_div & SDMMC_CLKDIV_CLK_DIVR0));

    /*
	 * クロック・イネーブル
	 */
	sil_wrw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_CLKENA), low_power_mode | SDMMC_CLKENA_CCLK_EN);
	return sdmmc_sendcommand(hsd, MCI_CLK, 0x0, NULL);
}


/*
 *  SDカード有無チェック
 */
bool_t
sdmmc_sense(int id)
{
	SDMMC_Handle_t *hsd = &SDHandle;

	if(hsd->status == SDMMC_STATUS_NONE)
		return false;
	if((sil_rew_mem((uint32_t *)(TADR_SDMMC1_BASE+TOFF_SDMMC_STAT)) & SDMMC_STAT_DATA_3_STAT) != 0)
		return true;
	else if(hsd->status == SDMMC_STATUS_ACTIVATE)
		return true;
	else
		return false;
}

/*
 *  SDMMC初期化
 *  parameter1  未使用
 */
void
sdmmc_init(intptr_t exinf)
{
	memset(&SDHandle, 0, sizeof(SDMMC_Handle_t));
    if(!sdmmc_clk_is_enabled()){
		syslog_0(LOG_ALERT, "SDMMC:sdmmc_init: no clock !");
		return;
    }
	SDHandle.base   = TADR_SDMMC1_BASE;
	SDHandle.status = SDMMC_STATUS_INIT;
}

/*
 *  SDMMC処理終了
 */
void
sdmmc_deinit(int id)
{
    /*
	 *  SD/MMCモジュールリセット
	 */
	sil_orw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_PERMODRST), RSTMGR_PERMODRST_SDMMC);
	SDHandle.status = SDMMC_STATUS_NONE;
}

/*
 *  SDMMCライトプロテクト判定
 *  parameter1  SDカードハンドラ
 *  result      if true, write protected.
 */
bool_t
sdmmc_card_is_write_protected(SDMMC_Handle_t *hsd)
{
	if((sil_rew_mem((uint32_t *)(hsd->base+TOFF_SDMMC_WRTPRT)) & SDMMC_WRTPRT_WR_PROTECT) != 0)
		return true;
	else
		return false;
}

/*
 *  SD/MMCバス幅設定
*/
static void
sdmmc_bus_width_set(SDMMC_Handle_t *hsd, uint16_t width)
{
	switch (width){
	case SDMMC_BUS_WIDE_8B:
		sil_wrw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_CTYPE), SDMMC_CTYPE_CARD_WIDTH1);
		break;
	case SDMMC_BUS_WIDE_4B:
		sil_wrw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_CTYPE), SDMMC_CTYPE_CARD_WIDTH2);
		break;
	case SDMMC_BUS_WIDE_1B:
		sil_wrw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_CTYPE), 0x0);
		break;
	default:
		break;
    }
}

/*
 *  ブロックサイズ取得
 */
static uint16_t
sdmmc_block_size_get(SDMMC_Handle_t *hsd)
{
	return (sil_rew_mem((uint32_t *)(hsd->base+TOFF_SDMMC_BLKSIZ)) & SDMMC_BLKSIZ_BLOCK_SIZE);
}

/*
 *  ブロックサイズ設定
 */
static void
sdmmc_block_size_set(SDMMC_Handle_t *hsd, uint16_t block_size)
{
	sil_wrw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_BLKSIZ), (block_size & SDMMC_BLKSIZ_BLOCK_SIZE));
}

/*
 *  転送バイト数設定
 */
static void
sdmmc_byte_count_set(SDMMC_Handle_t *hsd, uint32_t count)
{
	sil_wrw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_BYTCNT), count);
}

/*
 *  FIFOリセット
 */
static ER
sdmmc_fifo_reset(SDMMC_Handle_t *hsd)
{
    uint32_t timeout = SDMMC_MAX_T_POLL_COUNT;

	/*
	 *  リセット要求
	 */
	sil_orw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_CTL), SDMMC_CTL_FIFO_RST);

    /*
	 *  リセット終了待ち
	 */
	while((sil_rew_mem((uint32_t *)(hsd->base+TOFF_SDMMC_CTL)) & SDMMC_CTL_FIFO_RST) != 0){
		timeout--;
		if(timeout == 0)
			return E_TMOUT;
	}
	return E_OK;
}

/*
 *  使用中のFIFOカウントを返す
 */
static int32_t
sdmmc_fifo_count(SDMMC_Handle_t *hsd)
{
	int32_t fifocount = sil_rew_mem((uint32_t *)(hsd->base+TOFF_SDMMC_STAT));
	return (fifocount & SDMMC_STAT_FIFO_COUNT) >> 17;
}


/*
 *  ショートレスポンスを取り出す
 */
static uint32_t
sdmmc_read_short_response(SDMMC_Handle_t *hsd)
{
	return sil_rew_mem((uint32_t *)(hsd->base+TOFF_SDMMC_RESP0));
}

/*
 *  ロングレスポンスを取り出す
 */
static void
sdmmc_read_long_response(SDMMC_Handle_t *hsd, uint32_t *response)
{
	response[3] = sil_rew_mem((uint32_t *)(hsd->base+TOFF_SDMMC_RESP0));
	response[2] = sil_rew_mem((uint32_t *)(hsd->base+TOFF_SDMMC_RESP1));
	response[1] = sil_rew_mem((uint32_t *)(hsd->base+TOFF_SDMMC_RESP2));
	response[0] = sil_rew_mem((uint32_t *)(hsd->base+TOFF_SDMMC_RESP3));
}

/*
 *  読み込みスレッシュホールド設定
 */
static void
sdmmc_card_rd_threshold_enable(SDMMC_Handle_t *hsd, uint32_t threshold)
{
	uint32_t temp = (threshold << 16) & SDMMC_CARDTHRCTL_CARDRDTHRESHOLD;
	sil_wrw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_CARDTHRCTL), (temp | SDMMC_CARDTHRCTL_CARDRDTHREN));
}

/*
 *  SRC読み込み
 */
static ER
sdmmc_card_scr_get(SDMMC_Handle_t *hsd, uint32_t* scr_reg)
{
    ER ercd = E_OK;
    uint16_t prev_blk_size = sdmmc_block_size_get(hsd);

	sdmmc_block_size_set(hsd, SCR_LENGTH);
	sdmmc_byte_count_set(hsd, SCR_LENGTH);

	/*
	 *  FIFOリセット
	 */
    ercd = sdmmc_fifo_reset(hsd);
	if(ercd != E_OK)
		return ercd;

#ifndef SDMMC_READ_INNER_DMA
    sdmmc_card_rd_threshold_enable(hsd, SCR_LENGTH/4);
	hsd->pBuffPtr  = sbuffer;
	hsd->XferCount = SCR_LENGTH;
#else
	/*
	 *  DMA設定
	 */
    ercd = sdmmc_dma_reset(hsd);
	if(ercd != E_OK)
        return ercd;
    sdmmc_desc_chain_clear(hsd);
	sdmmc_dma_start(hsd, 0x0, SDMMC_DMA_PBL_1, false);
    sdmmc_dma_int_enable(hsd, SDMMC_DMA_INT_STATUS_ALL);
	ercd = sdmmc_setup_dma_trans(hsd, (uint32_t*)sbuffer, SCR_LENGTH);
	if(ercd != E_OK)
        return ercd;
#endif

    /*
	 *  CMD55送信(アペンド要求)
	 */
	ercd = sdmmc_sendcommand(hsd, MCI_CMD55, (hsd->RCA << 16), sdmmc_checkrep1);
	if(ercd != E_OK)
		return ercd;

    /*
	 *  ACMD51(転送要求)送信
	 */
	ercd = sdmmc_sendcommand(hsd, MCI_ACMD51, 0, sdmmc_checkrep1);
	if(ercd != E_OK)
		return ercd;

	/*
	 *  データ転送終了待ち
	 */
    ercd = sdmmc_wait_transfar(hsd, SDMMC_CMDTIMEOUT);
	if(ercd != E_OK)
		return ercd;
	memcpy(scr_reg, sbuffer, SCR_LENGTH);

	/*
	 *  ブロックサイズをリストアー
	 */
    sdmmc_block_size_set(hsd, prev_blk_size);
	return ercd;
}

/*
 *  ブロックサイズ設定
 */
static ER
sdmmc_card_block_size_set(SDMMC_Handle_t *hsd, const uint16_t blocksize)
{
	ER ercd = E_OK;
	/*
	 *  CMD16:ブロックサイズ設定
	 */
	ercd = sdmmc_sendcommand(hsd, MCI_CMD16, (uint32_t)blocksize, sdmmc_checkrep1);
    if(ercd != E_OK)
		return ercd;

	/*
	 *  コントローラブロックサイズ設定
	 */
	sdmmc_block_size_set(hsd, blocksize);
	return ercd;
}

/*
 *  スイッチファンクション
 */
static ER
sdmmc_card_switch_function(SDMMC_Handle_t *hsd, uint32_t function_arg, uint8_t * switch_status)
{
	ER ercd = E_OK;
	uint16_t prev_blk_size = sdmmc_block_size_get(hsd);

	/*
	 *  ブロックサイズとバイトカウントを64に設定
	 */
	sdmmc_block_size_set(hsd, SYSTEM_STATUS_LENGTH);
	sdmmc_byte_count_set(hsd, SYSTEM_STATUS_LENGTH);

	/*
	 *  FIFOをリセット
	 */
	ercd = sdmmc_fifo_reset(hsd);
	if(ercd != E_OK)
		return ercd;

#ifndef SDMMC_READ_INNER_DMA
    sdmmc_card_rd_threshold_enable(hsd, SYSTEM_STATUS_LENGTH/4);
	hsd->pBuffPtr  = sbuffer;
	hsd->XferCount = SYSTEM_STATUS_LENGTH;
#else
	/*
	 *  DMA設定
	 */
	ercd = sdmmc_dma_reset(hsd);
	if(ercd != E_OK)
		return ercd;
    sdmmc_desc_chain_clear(hsd);
    sdmmc_dma_start(hsd, 0x0, SDMMC_DMA_PBL_1, false);
    sdmmc_dma_int_enable(hsd, SDMMC_DMA_INT_STATUS_ALL);
	ercd = sdmmc_setup_dma_trans(hsd, (uint32_t*)switch_status, SYSTEM_STATUS_LENGTH);
	if(ercd != E_OK)
		return ercd;
#endif

	/*
	 *  CMD6送信
	 */
	ercd = sdmmc_sendcommand(hsd, MCI_CMD6, function_arg, sdmmc_checkrep1);
	if(ercd != E_OK)
		return ercd;

	/*
	 *  転送終了待ち
	 */
    ercd = sdmmc_wait_transfar(hsd, SDMMC_CMDTIMEOUT);
    sdmmc_block_size_set(hsd, prev_blk_size);
	return ercd;
}

/*
 *  SDMMCオープン
 */
SDMMC_Handle_t *
sdmmc_open(int id)
{
	ER ercd = E_OK;
	SDMMC_Handle_t *hsd = &SDHandle;
	volatile uint32_t timeout = SDMMC_RESET_TMO_INIT;
	uint32_t validvoltage = 0;
    uint32_t response = 0;
	uint32_t sdtype = SD_STD_CAPACITY;
    uint32_t clk_div;

    if(sdmmc_clk_is_enabled() != true)
		return NULL;

	clk_div = sdmmc_clock();
	if(clk_div == 0)
		return NULL;
	clk_div = clk_div / (4 * 2 * SDMMC_TRANSFAR_SPEED);

	hsd->base       = TADR_SDMMC1_BASE;
    hsd->cardtype   = SD_CARD_NOTDETECT;
	hsd->BusWide    = SDMMC_BUS_WIDE_4B;
	hsd->RetryCount = 32;
	hsd->semid      = SDMMC_SEM;
	hsd->ErrorCode  = 0;
	hsd->RCA        = 0;
	hsd->istatus    = 0;
	hsd->SdCmd      = 0;
	hsd->dma_enable = false;

    /*
	 *  SDカード電源オフ
	 */
    if(sdmmc_card_pwr_is_on(hsd)){
        sdmmc_card_pwr_off(hsd);
    }

    /*
	 *  SD/MMCモジュールリセット
	 */
	sil_orw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_PERMODRST), RSTMGR_PERMODRST_SDMMC);
    /*  Wait while SD/MMC module is reseting*/
    while (timeout--)
        ;
	sil_andw_mem((uint32_t *)(TADR_RSTMGR_BASE+TOFF_RSTMGR_PERMODRST), RSTMGR_PERMODRST_SDMMC);

	/*
	 *  カード・パワーオン
	 */
	sil_orw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_PWREN), SDMMC_PWREN_POWER_EN);

#ifndef SDMMC_READ_INNER_DMA
	/*
	 *  SD/MMC DMAを無効に設定
	 */
	sdmmc_dma_disable(hsd);
#else
	/*
	 *  SD/MMC DMAを有効に設定
	 */
    sdmmc_desc_chain_init(hsd);
	sdmmc_dma_enable(hsd);
#endif

    /* Enable default clock for alt_sdmmc_command_send to work*/
	sil_wrw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_CLKENA), SDMMC_CLKENA_CCLK_EN);
	ercd = sdmmc_sendcommand(hsd, MCI_CLK, 0x0, NULL);

    /*
	 *  バスを１に設定
	 */
    sdmmc_bus_width_set(hsd, SDMMC_BUS_WIDE_1B);

    ercd = sdmmc_card_clk_div_set(hsd, clk_div);
	if(ercd != E_OK){
		syslog_0(LOG_ERROR, "SDMMC:sdmmc_card_clk_div_set failed");
		return NULL;
    }

	/*
	 *  CMD0: GO IDLE STATE
	 */
	ercd = sdmmc_sendcommand(hsd, MCI_CMD0, 0x0, NULL);
	if(ercd != E_OK)
		return NULL;
	ercd = sdmmc_cmd_wait(hsd);
	if(ercd != E_OK){
		syslog_0(LOG_ERROR, "SDMMC:open go command failed");
		return NULL;
	}

	/*
	 *  CMD8: SEND IF COND
	 */
	ercd = sdmmc_sendcommand(hsd, MCI_CMD8, 0x1AA, sdmmc_checkrep1);
	if(ercd != E_OK)
		hsd->ErrorCode = SDMMC_ERROR_NONE;
	else{
        response = sdmmc_read_short_response(hsd);
		if(response == 0x1AA){
            hsd->cardtype = SD_CARD_V20;
			sdtype        = SD_HIGH_CAPACITY;
		}
	}

    timeout = SD_MAX_VOLT_TRIAL;
	/*
	 * CMD55を送信
	 */
	ercd = sdmmc_sendcommand(hsd, MCI_CMD55, 0, sdmmc_checkrep1);
	if(ercd == E_OK){
		/*
		 *  SDCARD確定:VOLTAGEの確認
		 */
		while(validvoltage == 0 && timeout > 0){
			/* SEND CMD55 APP_CMD with RCA as 0 */
			ercd = sdmmc_sendcommand(hsd, MCI_CMD55, 0, sdmmc_checkrep1);
			if(ercd != E_OK){
				return NULL;
			}
			/*
			 * 引数0x8010000でCMD41を送信
			 */
			ercd = sdmmc_sendcommand(hsd, MCI_ACMD41, (0x80100000 | sdtype), sdmmc_checkrep3);
			if(ercd != E_OK){
				return NULL;
			}
			/*
			 * レスポンスから制御ボルテージを取得
			 */
			response = sdmmc_read_short_response(hsd);
			validvoltage = (((response >> 31) == 1) ? 1 : 0);
			timeout--;
		}
		if(timeout == 0){
			syslog_0(LOG_ERROR, "SDMMC:open voltale timeout failed");
			return NULL;
		}

		if((response & SD_HIGH_CAPACITY) == SD_HIGH_CAPACITY){
			/*
			 *  HC-SDCARD確定
			 */
			hsd->cardtype = SD_CARD_HC;
		}
	}
	if(ercd != E_OK)
		return NULL;
	hsd->status = SDMMC_STATUS_READY;
	return hsd;
}

/*
 *  SDMMCクローズ
 */
ER
sdmmc_close(SDMMC_Handle_t *hsd)
{
	if(hsd == NULL)
		return E_PAR;

#ifdef SDMMC_READ_INNER_DMA
	/*
	 *  DMAディスクリプタクリア
	 */
    sdmmc_desc_chain_clear(hsd);
#endif

	/*
	 *  SDMMC割込み禁止
	 */
	sil_andw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_INTMSK), SDMMC_INT_STATUS_ALL);
	sdmmc_dma_disable(hsd);

    /*
	 *  Card電源オフ
	 */
    sdmmc_card_pwr_off(hsd);
	hsd->status = SDMMC_STATUS_INIT;
    return E_OK;
}

/*
 *  CIDの取得
 *  SDカードCMD2,MMCカードCMD1送信後のレスポンス受信
 *  parameter1  hsd: SDMMCハンドラ
 *  return      ERコード
 */
ER
sdmmc_checkCID(SDMMC_Handle_t *hsd)
{
	ER ercd  = E_OK;
	uint32_t i;

	if(hsd->cardtype == SD_IO_CARD)
		return E_OK;
	for(i = 0 ; i < hsd->RetryCount ; i++){
	    /*
		 *  CMD2送信
		 */
		ercd = sdmmc_sendcommand(hsd, MCI_CMD2, 0, sdmmc_checkrep2);
		if(ercd == E_OK){
		    sdmmc_read_long_response(hsd, &hsd->CID[0]);
			return E_OK;
		}
		sil_dly_nse(2000);
	}
	return ercd;
}

/*
 *  相対アドレス(RCA)の取得
 *  parameter1  hsd: SDMMCハンドラ
 *  return      ERコード
 */
ER
sdmmc_setaddress(SDMMC_Handle_t *hsd)
{
	ER ercd = E_OK;
	int i;
	if(hsd->cardtype == SD_CARD_V11 || hsd->cardtype == SD_CARD_V20 ||
		hsd->cardtype == SD_IO_COMBO_CARD || hsd->cardtype == SD_CARD_HC){

		for(i = 0 ; i < hsd->RetryCount ; i++){
			/*
			 *  CMD3送信、RCA受信
			 */
			ercd = sdmmc_sendcommand(hsd, MCI_CMD3, 0, sdmmc_checkrep6);
			if(ercd == E_OK){
			    /*
				 *  CMD10　CID設定
				 */
				ercd = sdmmc_sendcommand(hsd, MCI_CMD10, (hsd->RCA << 16), sdmmc_checkrep1);
				return E_OK;
			}
			sil_dly_nse(2000);
		}
	}
    hsd->RCA = 1;
	return ercd;
}

/*
 *  CSD(Card Specific DATA)の取得
 *  parameter1  hsd: SDMMCハンドラ
 *  return      ERコード
 */
ER
sdmmc_sendCSD(SDMMC_Handle_t *hsd)
{
	ER ercd = E_OK;
	int i;

	if(hsd->cardtype == SD_IO_CARD)
		return E_OK;
	for(i = 0 ; i < hsd->RetryCount ; i++){
		/*
		 *  CMD9送信、CSD要求
		 */
		ercd = sdmmc_sendcommand(hsd, MCI_CMD9, (uint32_t)(hsd->RCA << 16), sdmmc_checkrep2);
		if(ercd == E_OK){
			/*
			 *  CSD取得
			 */
		    sdmmc_read_long_response(hsd, &hsd->CSD[0]);
			return E_OK;
		}
		sil_dly_nse(2000);
	}
    return ercd;
}

/*
 *  SDカードの情報を取り込む
 *  parameter1  hsd: SDMMCハンドラ
 *  parameter2  pCardInfo: カードインフォ構造体へのポインタ
 *  return      ERコード
 */
ER
sdmmc_getcardinfo(SDMMC_Handle_t *hsd, SDMMC_CardInfo_t *pCardInfo)
{
	uint32_t rate_unit;
	uint32_t time_val;
	ER ercd = E_OK;
	uint32_t devicesize;
	uint8_t  blocklen, mul;

	hsd->CCC  = SDMMC_CSD_CCC_GET(hsd->CSD[1]);
    rate_unit = SDMMC_CSD_SPEED_RATE_GET(hsd->CSD[0]);
    time_val  = SDMMC_CSD_SPEED_TIME_GET(hsd->CSD[0]);
	if(time_val == 0 || rate_unit > 3)
		return E_SYS;

	pCardInfo->cardtype = (uint8_t)(hsd->cardtype);
	pCardInfo->RCA      = (uint16_t)(hsd->RCA);

	if(hsd->cardtype == SD_CARD_V11 || hsd->cardtype == SD_CARD_V20){
		/* ブロック長を取り込む */
		blocklen       = (uint8_t)((hsd->CSD[1] & 0x000F0000) >> 16);
		devicesize = (hsd->CSD[1] & 0x000003FF) << 2;
		devicesize |= (hsd->CSD[2] & 0xC0000000) >> 30;
		mul = (hsd->CSD[2] & 0x00038000) >> 15;
		pCardInfo->capacity  = (devicesize + 1) ;
		pCardInfo->capacity *= (1 << (mul + 2));
		pCardInfo->blocksize = 1 << (blocklen);
		pCardInfo->capacity *= pCardInfo->blocksize;
	}
	else if(hsd->cardtype == SD_CARD_HC){
		devicesize = (hsd->CSD[1] & 0x0000003F) << 16;
		devicesize |= (hsd->CSD[2] & 0xFFFF0000) >> 16;
		pCardInfo->capacity  = (uint64_t)(devicesize + 1) << 19 /* 512 * 1024*/;
		pCardInfo->blocksize = SDMMC_BLOCKSIZE;
	}
	else{
		pCardInfo->capacity  = 0;
		pCardInfo->blocksize = SDMMC_BLOCKSIZE;
		ercd = E_OBJ;
	}

	pCardInfo->maxsector = pCardInfo->capacity / SDMMC_BLOCKSIZE;
	pCardInfo->status    = (uint8_t)((hsd->CSD[3] >> 8) & 0xff);
	hsd->block_size      = pCardInfo->blocksize;
#if 1	/* ROI DEBUG */
	syslog_1(LOG_NOTICE, "## status[%02x] ##", pCardInfo->status);
	syslog_5(LOG_NOTICE, "## cardtype(%d) capacity[%08x%08x] blocksize(%d) maxsector(%u) ##", pCardInfo->cardtype, (uint32_t)(pCardInfo->capacity>>32), (uint32_t)(pCardInfo->capacity & 0xffffffff), pCardInfo->blocksize, pCardInfo->maxsector);
#endif	/* ROI DEBUG */
	return ercd;
}

/*
 *  SELECT_CARDコマンドの送信
 *  parameter1  hsd: SDMMCハンドラ
 *  return      ERコード
 */
ER
sdmmc_select_card(SDMMC_Handle_t *hsd, uint64_t addr)
{
	ER  ercd = E_OK;
	int i;

	for(i = 0 ; i < hsd->RetryCount ; i++){
		/*
		 *  CMD7送信 カードセレクト
		 */
		ercd = sdmmc_sendcommand(hsd, MCI_CMD7, (uint32_t)addr, sdmmc_checkrep1);
		if(ercd == E_OK)
			return E_OK;
		sil_dly_nse(2000);
	}
	return ercd;
}

/*
 *  SDMMCコンフィギュレーション
 *  parameter1  hsd: SDMMCハンドラ
 *  return      ERコード
 */
ER
sdmmc_configuration(SDMMC_Handle_t *hsd)
{
	ER ercd = E_OK;
	int i;

	for(i = 0 ; i < hsd->RetryCount ; i++){
		/*
		 *  CMD13送信 ステータス設定
		 */
		ercd = sdmmc_sendcommand(hsd, MCI_CMD13, (hsd->RCA << 16), sdmmc_checkrep1);
	    if(ercd == E_OK)
			break;
		sil_dly_nse(2000);
    }
	if(ercd != E_OK)
		return ercd;

	/*
	 *  SRC読み込み
	 */
	ercd = sdmmc_card_scr_get(hsd, &hsd->SCR[0]);
	if(ercd != E_OK)
		return ercd;

	/*
	 *  CCC確認
	 */
	hsd->high_speed = false;
	if(!(hsd->CCC & CCC_CLASS_10))	/*  version 1.01 or ! Class 10*/
        return E_OK;

	/*
	 *  スイッチステータス取得
	 */
	ercd = sdmmc_card_switch_function(hsd, SDMMC_SWITCH_FUNC_MODE_CHECK, sbuffer);
	if(ercd == E_OK){
		if(sbuffer[13] & SDMMC_SD_MODE_HIGH_SPEED){
			syslog_0(LOG_INFO, "SDMMC:High speed supported. Switching...");
			hsd->high_speed = true;
		}
	}
	return E_OK;
}

/*
 *  WIDE BUS設定
 *  parameter1  hsd: SDMMCハンドラ
 *  return      ERコード
 */
ER
sdmmc_set_widebus(SDMMC_Handle_t *hsd)
{
	uint32_t set_width_arg;
	ER ercd = E_OK;

	if((hsd->SCR[0] & (hsd->BusWide<<8)) == 0){
		hsd->BusWide = SDMMC_BUS_WIDE_1B;
	}
	syslog_1(LOG_INFO, "sdmmc_card_bus_width_set(%d) !", hsd->BusWide);

	switch (hsd->BusWide){
	case SDMMC_BUS_WIDE_8B:
		set_width_arg = 0x3;
		break;
	case SDMMC_BUS_WIDE_4B:
		set_width_arg = 0x2;
		break;
	case SDMMC_BUS_WIDE_1B:
		set_width_arg = 0x0;
		break;
	default:
		return E_PAR;
    }

	/*
	 *  CMD55送信(APPEDコマンド要求
	 */
	ercd = sdmmc_sendcommand(hsd, MCI_CMD55, (hsd->RCA << 16), sdmmc_checkrep1);
    if(ercd != E_OK)
		return ercd;

	/*
	 *  WIDE BUS設定でACMD6コマンド送信
	 */
	ercd = sdmmc_sendcommand(hsd, MCI_ACMD6, set_width_arg, sdmmc_checkrep1);
    if(ercd != E_OK)
		return ercd;
	/*
	 *  コントローラ用バス幅設定
	 */
    sdmmc_bus_width_set(hsd, hsd->BusWide);
    return ercd;
}

/*
 *  データREAD/WRITE転送
 */
static ER
sdmmc_transfer(SDMMC_Handle_t *hsd, uint64_t start_addr, uint32_t buffer[], const size_t buf_len)
{
	ER ercd = E_OK;
	uint32_t block_count;
	uint32_t byte_count;

	ercd = sdmmc_card_block_size_set(hsd, hsd->block_size);
	if(ercd != E_OK)
		return ercd;

    /*
	 *  転送ブロックカウント、バイトカウント算出
	 */
    block_count = buf_len / hsd->block_size;
    byte_count = block_count * hsd->block_size;

    /*  reset FIFO*/
    ercd = sdmmc_fifo_reset(hsd);
	if(ercd != E_OK)
        return ercd;

	sdmmc_byte_count_set(hsd, byte_count);
    sdmmc_card_rd_threshold_enable(hsd, SDMMC_BLOCKSIZE/4);

#ifdef SDMMC_READ_INNER_DMA
	if(hsd->dma_enable){	/* DMA使用設定 */
		ercd = sdmmc_dma_reset(hsd);
		if(ercd != E_OK)
			return ercd;
		sdmmc_desc_chain_clear(hsd);
		sdmmc_dma_start(hsd, 0x0, SDMMC_DMA_PBL_1, false);
		sdmmc_dma_int_enable(hsd, SDMMC_DMA_INT_STATUS_ALL);
	    ercd = sdmmc_setup_dma_trans(hsd, buffer, byte_count);
	}
	else{					/* 割込みFIFO使用設定 */
#endif
		hsd->pBuffPtr  = (uint8_t *)buffer;
		hsd->XferCount = byte_count;
#ifdef SDMMC_READ_INNER_DMA
	}
#endif
	if(ercd != E_OK)
		return ercd;

    /*
	 *  READ/WRITEコマンド送信
	 */
    if(hsd->cardtype == SD_CARD_HC)
		ercd = sdmmc_sendcommand(hsd, hsd->SdCmd, start_addr / hsd->block_size, sdmmc_checkrep1);
	else
		ercd = sdmmc_sendcommand(hsd, hsd->SdCmd, start_addr, sdmmc_checkrep1);
	return ercd;
}

/*
 *  SDMMCブロックREAD
 *  parameter1  hsd: SDMMCハンドラ
 *  parameter2  pbuf: 読み出しデータ格納領域のポインタ(uint32_t型)
 *  parameter3  ReadAddr: カード上の読み出し位置
 *  parameter4  blocksize: ブロックサイズ
 *  parameter5  num: 読み出しブロック数
 *  return      ERコード
 */
ER
sdmmc_blockread(SDMMC_Handle_t *hsd, void *dest, uint64_t ReadAddr, uint32_t blocksize, uint32_t num)
{
	ER ercd = E_OK;
	uint32_t tnum;
	uint32_t cnum = num;

	if(hsd == NULL || num == 0)
		return E_PAR;
	if(blocksize != hsd->block_size)
		return E_PAR;
	if((ReadAddr % hsd->block_size) != 0)
		return E_PAR;
	if(hsd->status != SDMMC_STATUS_READY)
		return E_OBJ;
	if(!sdmmc_is_idle(hsd))
		return E_SYS;
	hsd->status = SDMMC_STATUS_ACTIVATE;
	if(num == 1)
		hsd->SdCmd = MCI_CMD17;
	else
		hsd->SdCmd = MCI_CMD18;
	sdmmc_dma_enable(hsd);
	while(cnum > 0){
		tnum = MIN(SDMMC_DESC_COUNT, cnum);
		ercd = sdmmc_transfer(hsd, ReadAddr, dest, blocksize * tnum);
		if(ercd != E_OK)
			break;
		ReadAddr += blocksize * tnum;
		cnum -= tnum;
		if(cnum > 0)
			ercd = sdmmc_wait_transfar(hsd, SDMMC_CMDTIMEOUT);
		if(ercd != E_OK)
			break;
	}
	if(ercd != E_OK)
		hsd->status = SDMMC_STATUS_READY;
	return ercd;
}

/*
 *  SDMMCブロックWRITE
 *  parameter1  hsd: SDMMCハンドラ
 *  parameter2  pbuf: 書込みデータ格納領域のポインタ(uint32_t型)
 *  parameter3  WritedAddr: カード上の書き込み位置
 *  parameter4  blocksize: ブロックサイズ
 *  parameter5  num: 書込みブロック数
 *  return      ERコード
 */
ER
sdmmc_blockwrite(SDMMC_Handle_t *hsd, void *src, uint64_t WriteAddr, uint32_t blocksize, uint32_t num)
{
	ER ercd = E_OK;

	if(hsd == NULL || num == 0)
		return E_PAR;
	if(blocksize != hsd->block_size)
		return E_PAR;
	if((WriteAddr % hsd->block_size) != 0)
		return E_PAR;
	if(hsd->status != SDMMC_STATUS_READY)
		return E_OBJ;
	if(!sdmmc_is_idle(hsd))
		return E_SYS;
	hsd->status = SDMMC_STATUS_ACTIVATE;
	if(num == 1)
		hsd->SdCmd = MCI_CMD24;
	else
		hsd->SdCmd = MCI_CMD25;
	sdmmc_dma_enable(hsd);
	ercd = sdmmc_transfer(hsd, WriteAddr, src, blocksize * num);
	if(ercd != E_OK)
		hsd->status = SDMMC_STATUS_READY;
	return ercd;
}

/*
 *  SDMMCブロック転送終了待ち
 *  parameter1  hsd: SDMMCハンドラ
 *  parameter2  Timeout: タイムアウト値(1msec)
 *  return      ERコード
 */
ER
sdmmc_wait_transfar(SDMMC_Handle_t *hsd, uint32_t Timeout)
{
	ER ercd = E_TMOUT;
#ifdef SDMMC_READ_INNER_DMA
	uint32_t idmac_status;
#endif

	if(hsd == NULL)
		return E_PAR;
	if(hsd->status < SDMMC_STATUS_READY)
		return E_OBJ;
	while (--Timeout){
		/*
		 *  エラー発生チェック
		 */
		if(hsd->ErrorCode != SDMMC_ERROR_NONE){
			ercd = E_SYS;
			break;
		}

#ifdef SDMMC_READ_INNER_DMA
		/*
		 *  DMA転送チェック
		 */
	    idmac_status = hsd->dstatus;
        if(idmac_status & SDMMC_IDSTS_AIS){
			hsd->dstatus &= ~SDMMC_IDSTS_AIS;
			ercd = E_SYS;
            syslog_0(LOG_ERROR, "SDMMC:Abnormal DMA reported.");
            break;
		}
#endif

		/*
		 *  SDMMC転送チェック
		 */
		if(hsd->XferCount == 0){
			ercd = E_OK;
			break;
		}
		twai_sem(hsd->semid, 1);
	}

	Timeout = SDMMC_TMO_WAITER;
	while(!sdmmc_is_idle(hsd) && --Timeout){
		dly_tsk(1);
	}
	if(Timeout == 0){
		syslog_0(LOG_ERROR, "SDMMC:Timed out waiting for SDMMC to become idle");
		ercd = E_TMOUT;
	}
	if(hsd->SdCmd == MCI_CMD18 || hsd->SdCmd == MCI_CMD25){
		/*
		 *  マルチブロック転送ならCMD12を送信
		 */
		ercd = sdmmc_sendcommand(hsd, MCI_CMD12, 0, sdmmc_checkrep1);
	}
	if((hsd->SdCmd & SDMMC_CMD_RD_WR) != 0)
		dly_tsk(SDMMC_WRITE_DELAY_TIME);
	hsd->status = SDMMC_STATUS_READY;
    return ercd;
}

/*
 *  SDMMC割込みハンドラ
 */
void
sdmmc_hps_handler(void)
{
	SDMMC_Handle_t *hsd = &SDHandle;
	uint32_t istatus, dstatus;
	istatus = sil_rew_mem((uint32_t *)(hsd->base+TOFF_SDMMC_MINTSTS));
	dstatus = sil_rew_mem((uint32_t *)(hsd->base+TOFF_SDMMC_IDSTS));
	uint32_t level, freespace, fullspace, temp;
	volatile uint8_t *p;
	int      i;

	syslog_2(LOG_DEBUG, "## sdmmc_hps_handler[%08x][%08x] ##", istatus, dstatus);
	if(istatus != 0){
		hsd->istatus   |= istatus;
		hsd->ErrorCode |= istatus & SDMMC_ERROR_MASK;
		if((istatus & SDMMC_INTSTS_TXDR) != 0){
			if(hsd->XferCount > 0 && hsd->pBuffPtr != NULL){
		        level = sdmmc_fifo_count(hsd);

		        /*
				 *  送信TX FIFO設定
				 */
	            freespace = SDMMC_FIFO_NUM_ENTRIES - level;
	            freespace = MIN(hsd->XferCount / 4, freespace);
	            for(i = 0; i < freespace; i++){
					p = (volatile uint8_t *)hsd->pBuffPtr;
					temp = (p[3]<<24) | (p[2]<< 16) | (p[1]<<8) | p[0];
					sil_wrw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_DATA), temp);
					hsd->pBuffPtr += 4;
	            }
	            hsd->XferCount -= freespace * 4;
			}
			else
				sil_andw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_INTMSK), SDMMC_INTSTS_TXDR);
		}
		if((hsd->istatus & (SDMMC_INTSTS_RXDR | SDMMC_INTSTS_DTO)) != 0){
#ifndef SDMMC_READ_INNER_DMA
	        level = sdmmc_fifo_count(hsd);
			fullspace = MIN(hsd->XferCount / 4, level);
			if(hsd->XferCount > 0 && hsd->pBuffPtr != NULL){
	            for(i = 0; i < fullspace; i++){
					p = (volatile uint8_t *)hsd->pBuffPtr;
					temp = sil_rew_mem((uint32_t *)(hsd->base+TOFF_SDMMC_DATA));
					p[0] = temp & 0xff;
					p[1] = (temp >> 8) & 0xff;
					p[2] = (temp >> 16) & 0xff;
					p[3] = (temp >> 24);
					hsd->pBuffPtr += 4;
	            }
	            hsd->XferCount -= fullspace * 4;
				level -= fullspace;
			}
			if(hsd->XferCount == 0 && level > 0){
				syslog_2(LOG_ERROR, "sdmmc_hps_handler level(%d) XferCount(%d) !", level, hsd->XferCount);
	            for(i = 0; i < level; i++){
					temp = sil_rew_mem((uint32_t *)(hsd->base+TOFF_SDMMC_DATA));
				}
			}
#else
			if(hsd->XferCount > 0 && hsd->pBuffPtr == NULL){
				hsd->istatus &= ~SDMMC_INTSTS_DTO;
				hsd->XferCount = 0;
			}
#endif
		}
		/*
		 *  SDMMC割込みステータスクリア
		 */
		sil_wrw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_RINTSTS), istatus);
	}
	if(dstatus != 0){
		hsd->dstatus |= dstatus;
		/*
		 *  SDMMC-DMA割込みステータスクリア
		 */
		sil_wrw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_IDSTS), dstatus);
	}
	isig_sem(hsd->semid);
}

/*
 * SDMMCコマンド送信
 */
static ER
sdmmc_sendcommand(SDMMC_Handle_t *hsd, uint32_t cmd, uint32_t arg, ER (*func)(SDMMC_Handle_t *, uint32_t))
{
    /*  Create interrupt mask by command configurations*/
    uint32_t int_mask = SDMMC_INTSTS_RE             /*  Response error*/
                      | SDMMC_INTSTS_BAR            /*  Response timeout*/
                      | SDMMC_INTSTS_CD             /*  Card detect (CD) interrupt*/
                      | SDMMC_INTSTS_HLE            /*  Hardware Locked Write Error*/
                      | SDMMC_INTSTS_CMD;           /*  Command done (CD) interrupt*/
    ER ercd = E_OK;
	uint32_t cmd_index = cmd & CMD_INDEX;
	uint32_t timeout = SDMMC_TMO_WAITER;

	int_mask |= SDMMC_INTSTS_DTO		/*  Data transfer over*/
			  | SDMMC_INTSTS_RCRC		/*  Response CRC error*/
			  | SDMMC_INTSTS_DCRC		/*  Data CRC error*/
			  | SDMMC_INTSTS_HTO		/*  Data starvation by host timeout*/
			  | SDMMC_INTSTS_FRUN		/*  FIFO underrun/overrun error*/
			  | SDMMC_INTSTS_EBE;		/*  End-bit error*/

	if((cmd & SDMMC_CMD_RD_WR) != 0){
		int_mask |= SDMMC_INTSTS_TXDR	/*  Transmit FIFO data request (TXDR)*/
				  | SDMMC_INTSTS_HLE;	/*  Hardware locked write error (HLE)*/
	}
	else{
		int_mask |= SDMMC_INTSTS_RXDR	/*  Receive FIFO data request (RXDR)*/
				  | SDMMC_INTSTS_SBE;	/*  Start-bit error (SBE)*/
	}

	if(cmd_index != CMD_INDEX){
        while (sdmmc_is_busy(hsd) && timeout--){
			dly_tsk(1);
		}
	}

	/*
	 *  SDMMC割込みステータスクリア
	 */
	hsd->istatus = 0;
	sil_wrw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_RINTSTS), SDMMC_INT_STATUS_ALL);

	/*
	 *  SDMMC割込み設定
	 */
	sil_orw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_CTL), SDMMC_CTL_INT_EN);
	sil_wrw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_INTMSK), int_mask);

    /*
	 *  コマンドと引数設定
	 */
	sil_wrw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_CMDARG), arg);
    sil_wrw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_CMD), cmd);
    sil_wrw_mem((uint32_t *)(hsd->base+TOFF_SDMMC_CMD), cmd | SDMMC_CMD_START_CMD);

	/*
	 *  クロック送出ケース
	 */
	if(cmd_index == CMD_INDEX){
		timeout = SDMMC_TMO_WAITER;
		ercd = E_TMOUT;
		while (--timeout){
			uint32_t cmd_register = sil_rew_mem((uint32_t *)(hsd->base+TOFF_SDMMC_CMD));
			/*
			 *  エラーチェック
			 */
	        if(hsd->ErrorCode != SDMMC_ERROR_NONE){
				ercd = E_SYS;
				break;
			}

			/*
			 *  コマンドスタートチェック
			 */
			if((cmd_register & SDMMC_CMD_START_CMD) == 0){
				ercd = E_OK;
				break;
			}
			dly_tsk(1);
		}
		return ercd;
    }

    /*
	 *  コマンド送信待ち
	 */
	if(func != NULL){
		ercd = sdmmc_cmd_wait(hsd);
		if(ercd == E_OK)
			func(hsd, cmd);
	}
	return ercd;
}

/*
 *  R1レスポンス
 */
static ER
sdmmc_checkrep1(SDMMC_Handle_t *hsd, uint32_t Cmd)
{
	uint32_t respValue;
	/*
	 *  レスポンス１のエラーチェック
	 */
	respValue = sdmmc_read_short_response(hsd);
	if((respValue & SD_OCR_ERROR) == 0)
		return E_OK;
	else
		return E_SDCOM;
}

/*
 *  R2レスポンス
 */
static ER
sdmmc_checkrep2(SDMMC_Handle_t *hsd, uint32_t Cmd)
{
	return E_OK;
}

/*
 *  R3レスポンス
 */
static ER
sdmmc_checkrep3(SDMMC_Handle_t *hsd, uint32_t Cmd)
{
	return E_OK;
}

/*
 *  R6レスポンス
 */
static ER
sdmmc_checkrep6(SDMMC_Handle_t *hsd, uint32_t Cmd)
{
	uint32_t respValue;

	/*
	 *  レスポンス１の確認
	 */
	respValue = sdmmc_read_short_response(hsd);
    hsd->RCA = (uint16_t) (respValue >> 16);
	return E_OK;
}

/*
 *  コマンド転生終了待ち
 */
static ER
sdmmc_cmd_wait(SDMMC_Handle_t *hsd)
{
	ER ercd = E_TMOUT;
    uint32_t timeout = SDMMC_CMDTIMEOUT;

	while(timeout > 0){
		/*
		 *  SDMMCエラーチェック
		 */
		if(hsd->ErrorCode != SDMMC_ERROR_NONE){
			ercd = E_SYS;
			break;
        }

        /*
		 *  コマンド転送終了待ち
		 */
		if(hsd->istatus & SDMMC_INTSTS_CMD){
			ercd = E_OK;
			break;
		}
		twai_sem(hsd->semid, 1);
		timeout--;
	}
	hsd->istatus &= ~SDMMC_INTSTS_CMD;
	return ercd;
}


